"""Tools for reading and writing geospatial data formats."""

import logging
import os
import tarfile
import tempfile
import warnings
import zipfile
from collections.abc import Sequence
from pathlib import Path
from re import search
from typing import Optional, Union

from . import gis_import_error_message


try:
    import fiona
    import rasterio
    from pyproj import CRS
    from shapely.geometry import shape
except (ImportError, ModuleNotFoundError) as err:
    msg = gis_import_error_message.format(Path(__file__).stem)
    raise ImportError(msg) from err

LOGGER = logging.getLogger("RavenPy")
WGS84 = 4326


# Function addressing exploit CVE-2007-4559
def is_within_directory(directory: Union[str, os.PathLike], target: Union[str, os.PathLike]) -> bool:
    abs_directory = Path(directory).resolve()
    abs_target = Path(target).resolve()

    prefix = os.path.commonprefix([abs_directory, abs_target])

    return prefix == abs_directory


# Function addressing exploit CVE-2007-4559
def safe_extract(tar: tarfile.TarFile, path: str = ".", members=None, *, numeric_owner=False) -> None:
    for member in tar.getmembers():
        member_path = Path(path).joinpath(member.name)
        if not is_within_directory(path, member_path):
            raise Exception("Attempted Path Traversal in Tar File")

    tar.extractall(path, members, numeric_owner=numeric_owner)  # noqa: S202


def address_append(address: Union[str, Path]) -> str:
    """
    Format a URL/URI to be more easily read with libraries such as "rasterstats".

    Parameters
    ----------
    address : Union[str, Path]
        URL/URI to a potential zip or tar file.

    Returns
    -------
    str
        URL/URI prefixed for archive type.
    """
    zipped = search(r"(\.zip)", str(address))
    tarred = search(r"(\.tar)", str(address))

    try:
        if zipped:
            return f"zip://{address}"
        elif tarred:
            return f"tar://{address}"
        else:
            LOGGER.info("No prefixes needed for address.")
            return str(address)
    except Exception:
        msg = "Failed to prefix or parse URL {address}."
        LOGGER.error(msg)
        raise


def generic_extract_archive(
    resources: Union[str, Path, list[Union[bytes, str, Path]]],
    output_dir: Optional[Union[str, Path]] = None,
) -> list[str]:
    """
    Extract archives (tar/zip) to a working directory.

    Parameters
    ----------
    resources : str or Path or list of bytes or str or Path
        List of archive files (if netCDF files are in list, they are passed and returned as well in the return).
    output_dir : str or Path, optional
        String or Path to a working location (default: temporary folder).

    Returns
    -------
    list of str
        A list of original or of extracted files.
    """
    archive_types = [".tar", ".zip", ".7z"]
    output_dir = output_dir or tempfile.gettempdir()

    if not isinstance(resources, list):
        resources = [resources]

    files = list()

    for arch in resources:
        if any(ext in str(arch).lower() for ext in archive_types):
            try:
                LOGGER.debug("archive=%s", arch)
                file = Path(arch).name

                if file.endswith(".nc"):
                    files.append(Path(output_dir.join(arch)))
                elif file.endswith(".tar"):
                    with tarfile.open(arch, mode="r") as tar:
                        safe_extract(tar, path=output_dir)  # Function addressing exploit CVE-2007-4559
                        files.extend([str(Path(output_dir).joinpath(f)) for f in tar.getnames()])
                elif file.endswith(".zip"):
                    with zipfile.ZipFile(arch, mode="r") as zf:
                        zf.extractall(path=output_dir)  # noqa: S202
                        files.extend([str(Path(output_dir).joinpath(f)) for f in zf.namelist()])
                elif file.endswith(".7z"):
                    msg = "7z file extraction is not supported at this time."
                    LOGGER.warning(msg)
                    warnings.warn(msg, UserWarning, stacklevel=2)
                else:
                    msg = f'File extension "{file}" unknown'
                    LOGGER.debug(msg)
            except OSError as e:
                msg = f"Failed to extract archive {arch}: {e}"
                LOGGER.error(msg)
        else:
            LOGGER.warning("No archives found. Continuing...")
            return resources

    return files


def archive_sniffer(
    archives: Union[str, Path, list[Union[str, Path]]],
    working_dir: Optional[Union[str, Path]] = None,
    extensions: Optional[Sequence[str]] = None,
) -> list[Union[str, Path]]:
    """
    Return a list of locally unarchived files that match the desired extensions.

    Parameters
    ----------
    archives : str or Path or list of str or Path
        Archive location or list of archive locations.
    working_dir : str or Path, optional
        String or Path to a working location.
    extensions : Sequence of str, optional
        List of accepted extensions.

    Returns
    -------
    list of str or Path
        A list of files with matching accepted extensions.
    """
    potential_files = list()

    if not extensions:
        extensions = [".gml", ".shp", ".geojson", ".gpkg", ".json"]

    decompressed_files = generic_extract_archive(archives, output_dir=working_dir)
    for file in decompressed_files:
        if any(ext in Path(file).suffix for ext in extensions):
            potential_files.append(file)  # noqa: PERF401
    return potential_files


def crs_sniffer(*args: Union[str, Path, Sequence[Union[str, Path]]]) -> Union[list[Union[str, int]], str, int]:
    r"""
    Return the list of CRS found in files.

    Parameters
    ----------
    \*args : str or Path or sequence of str or Path
        Path(s) to the file(s) to examine.

    Returns
    -------
    str or list of str
        Returns either a list of CRSes or a single CRS definition, depending on the number of instances found.
    """
    crs_list = list()
    vectors = (".gml", ".shp", ".geojson", ".gpkg", ".json")
    rasters = (".tif", ".tiff")
    all_files = vectors + rasters

    for file in args:
        found_crs = False
        suffix = Path(file).suffix.lower()
        try:
            if suffix == ".zip":
                file = archive_sniffer(file, extensions=all_files)[0]
                suffix = Path(file).suffix.lower()

            if suffix in vectors:
                if suffix == ".gpkg":
                    if len(fiona.listlayers(file)) > 1:
                        raise NotImplementedError
                with fiona.open(file, "r") as src:
                    found_crs = CRS.from_wkt(src.crs_wkt).to_epsg()
            elif suffix in rasters:
                with rasterio.open(file, "r") as src:
                    found_crs = CRS.from_user_input(src.crs).to_epsg()
            else:
                raise FileNotFoundError("Invalid filename suffix")
        except FileNotFoundError as e:
            msg = f"{e}: Unable to open file {args}"
            LOGGER.warning(msg)
            raise Exception(msg) from e
        except NotImplementedError as e:
            msg = f"{e}: Multilayer GeoPackages are currently unsupported"
            LOGGER.error(msg)
            raise Exception(msg) from e
        except RuntimeError as e:
            LOGGER.error(e)
            pass

        crs_list.append(found_crs)

    if crs_list is None:
        msg = f"No CRS definitions found in {args}."
        raise FileNotFoundError(msg)

    if len(crs_list) == 1:
        if not crs_list[0]:
            msg = f"No CRS definitions found in {args}. Assuming {WGS84}."
            LOGGER.warning(msg)
            warnings.warn(msg, UserWarning, stacklevel=2)
            return WGS84
        return crs_list[0]
    return crs_list


def raster_datatype_sniffer(file: Union[str, Path]) -> str:
    """
    Return the type of the raster stored in the file.

    Parameters
    ----------
    file : str or Path
        Path to file.

    Returns
    -------
    str
        A rasterio datatype of array values.
    """
    try:
        with rasterio.open(file, "r") as src:
            dtype = src.dtypes[0]
        return dtype
    except rasterio.errors.RasterioError as e:
        msg = f"Unable to read data type from file: {file}."
        LOGGER.exception(msg)
        raise ValueError(msg) from e


def get_bbox(vector: Union[str, Path], all_features: bool = True) -> tuple[float, float, float, float]:
    """
    Return bounding box of all features or the first feature in file.

    Parameters
    ----------
    vector : str or Path
        A path to file storing vector features.
    all_features : bool
        Return the bounding box for all features. Default: True.

    Returns
    -------
    float, float, float, float
        Geographic coordinates of the bounding box (lon0, lat0, lon1, lat1).
    """
    if not all_features:
        with fiona.open(vector, "r") as src:
            for feature in src:
                geom = shape(feature["geometry"])
                return geom.bounds

    with fiona.open(vector, "r") as src:
        return src.bounds



from collections.abc import Sequence
from textwrap import dedent, indent
from typing import Literal, Optional

from .base import Sym
from .commands import Command, Process


# --- Processes --- #
class Precipitation(Process):
    """Precipitation."""

    algo: Literal["PRECIP_RAVEN", "RAVEN_DEFAULT"] = "PRECIP_RAVEN"


class CapillaryRise(Process):
    algo: Literal["RISE_HBV", "CRISE_HBV"]


class CanopyEvaporation(Process):
    algo: Literal["CANEVP_RUTTER", "CANEVP_MAXIMUM", "CANEVP_ALL"]


class CanopySublimation(Process):
    algo: Literal["CANEVP_ALL", "CANEVP_MAXIMUM", "CANSUBLIM_ALL", "CANSUBLIM_MAXIMUM"]


class SoilBalance(Process):
    algo: Literal["SOILBAL_SACSMA"]


class SoilEvaporation(Process):
    algo: Literal[
        "SOILEVAP_VIC",
        "SOILEVAP_HBV",
        "SOILEVAP_HYPR",
        "SOILEVAL_CHU",
        "SOILEVAP_UBC",
        "SOILEVAP_GR4J",
        "SOILEVAP_TOPMODEL",
        "SOILEVAP_SEQUEN",
        "SOILEVAP_ROOT",
        "SOILEVAP_ROOT_CONSTRAIN",
        "SOILEVAP_ROOTFRAC",
        "SOILEVAP_GAWSER",
        "SOILEVAP_FEDERER",
        "SOILEVAP_ALL",
        "SOILEVAP_LINEAR",
        "SOILEVAP_SACSMA",
        "SOILEVAP_HYMOD2",
    ]


class LakeEvaporation(Process):
    algo: Literal["LAKE_EVAP_BASIC"]


class LakeFreeze(Process):
    algo: Literal["LFREEZE_BASIC", "LFREEZE_THERMAL"]


class OpenWaterEvaporation(Process):
    algo: Literal["OPEN_WATER_EVAP", "OPEN_WATER_RIPARIAN"]


class Infiltration(Process):
    algo: Literal[
        "INF_RATIONAL",
        "INF_SCS",
        "INF_ALL_INFILTRATES",
        "INF_GR4J",
        "INF_GREEN_AMPT",
        "INF_GA_SIMPLE",
        "INF_UPSCALED_GREEN_AMPT",
        "INF_HBV",
        "INF_UBC",
        "INF_VIC",
        "INF_VIC_ARNO",
        "INF_PRMS",
        "INF_HMETS",
    ]


class Percolation(Process):
    algo: Literal[
        "PERC_GAWSER",
        "PERC_LINEAR",
        "PERC_POWER_LAW",
        "PERC_PRMS",
        "PERC_SACRAMENTO",
        "PERC_CONSTANT",
        "PERC_GR4J",
        "PERC_GR4JEXCH",
        "PERC_GR4JEXCH2",
    ]


class CapilaryRise(Process):
    algo: Literal["CRISE_HBV"]


class Baseflow(Process):
    algo: Literal[
        "BASE_LINEAR",
        "BASE_LINEAR_CONSTRAIN",
        "BASE_LINEAR_ANALYTIC",
        "BASE_POWER_LAW",
        "BASE_CONSTANT",
        "BASE_VIC",
        "BASE_THRESH_POWER",
        "BASE_THRESH_STOR",
        "BASE_GR4J",
        "BASE_TOPMODEL",
        "BASE_SACRAMENTO",
    ]


class Interflow(Process):
    algo: Literal["INTERFLOW_PRMS"]


class Recharge(Process):
    algo: Literal[
        "RECHARGE_FROMFILE",
        "RECHARGE_CONSTANT",
        "RECHARGE_DEFAULT",
        "RECHARGE_CONSTANT_OVERLAP",
        "RECHARGE_DATA",
        "RECHARGE_FLUX",
        "RAVEN_DEFAULT",
    ]


class Seepage(Process):
    algo: Literal["SEEP_LINEAR"]


class DepressionOverflow(Process):
    algo: Literal["DFLOW_THRESHPOW", "DFLOW_LINEAR"]


class LakeRelease(Process):
    algo: Literal["LAKEREL_LINEAR"]


class Abstraction(Process):
    algo: Literal["ABST_PERCENTAGE", "ABST_FILL", "ABST_SCS", "ABST_PDMROF"]


class SnowMelt(Process):
    algo: Literal["MELT_POTMELT"]


class SnowRefreeze(Process):
    algo: Literal["FREEZE_DEGREE_DAY"]


class SnowBalance(Process):
    algo: Literal[
        "SNOBAL_SIMPLE_MELT",
        "SNOBAL_COLD_CONTENT",
        "SNOBAL_HBV",
        "SNOBAL_TWO_LAYER",
        # "SNOWBAL_CEMA_NEIGE",
        "SNOBAL_CEMA_NIEGE",
        "SNOBAL_HMETS",
        "SNOWBAL_GAWSER",
        "SNOWBAL_UBC",
    ]


class Sublimation(Process):
    algo: Literal[
        "SUBLIM_SVERDRUP",
        "SUBLIM_KUZMIN",
        "SUBLIM_CENTRAL_SIERRA",
        "SUBLIM_PSBM",
        "SUBLIM_WILLIAMS",
    ]


class SnowAlbedoEvolve(Process):
    algo: Literal["SNOALB_UBC"]


class SnowTempEvolve(Process):
    algo: Literal["SNOTEMP_NEWTONS"]


class CanopyDrip(Process):
    algo: Literal["CANDRIP_RUTTER", "CANDRIP_SLOWDRAIN"]


class CropHeatUnitEvolve(Process):
    algo: Literal["CHU_ONTARIO"]


class GlacierMelt(Process):
    algo: Literal["GMELT_SIMPLE_MELT", "GMELT_HBV", "GMELT_UBC"]


class GlacierRelease(Process):
    algo: Literal["GRELEASE_LINEAR", "GRELEASE_HBV_EC"]


class Flush(Process):
    algo: Literal["RAVEN_DEFAULT"] = "RAVEN_DEFAULT"
    p: Optional[float] = None


class Overflow(Process):
    algo: Literal["OVERFLOW_RAVEN", "RAVEN_DEFAULT"]
    _sub = "-->"
    _indent = "    "


class Split(Process):
    p: float = None
    to: tuple[str, str]


class Convolve(Process):
    algo: Literal["CONVOL_GR4J_1", "CONVOL_GR4J_2", "CONVOL_GAMMA", "CONVOL_GAMMA_2"]


class LateralFlush(Process):
    """Lateral flush."""


class LateralEquilibrate(Process):
    """
    Lateral equilibrate.

    Instantaneously equilibrates groundwater storage in basin HRUs.
    """


# --- End processes --- #


class Conditional(Command):
    """Conditional statement."""

    kind: Literal["HRU_TYPE", "LAND_CLASS", "HRU_GROUP"]
    op: Literal["IS", "IS_NOT"]
    value: str

    _sub = "-->"

    def to_rv(self):
        cmd = self._sub + self.__class__.__name__
        return f":{cmd:<20} {self.kind} {self.op} {self.value}"


class ProcessGroup(Command):
    p: Sequence[Process]
    params: Sequence[Sym]

    @property
    def _template(self):
        return """
           :{_cmd}
           {_processes}
           :End{_cmd} CALCULATE_WTS {_params}
           """

    @property
    def _indent(self):
        return "    "

    def to_rv(self):
        d = {
            "_cmd": "ProcessGroup",
            "_processes": indent("\n".join(str(p) for p in self.p), self._indent),
            "_params": " ".join(map(str, self.params)),
        }
        return dedent(self._template).format(**d)



from collections.abc import Sequence
from typing import Literal, Union

from pydantic import Field, field_validator
from pydantic.dataclasses import dataclass
from pymbolic.primitives import Variable

import ravenpy.config.processes as p
from ravenpy.config import commands as rc
from ravenpy.config import options as o
from ravenpy.config.base import Params, Sym, SymConfig
from ravenpy.config.commands import HRU, PL, Process
from ravenpy.config.defaults import nc_attrs
from ravenpy.config.rvs import Config


@dataclass(config=SymConfig)
class P(Params):
    GAMMA_SHAPE: Sym = Variable("GAMMA_SHAPE")
    GAMMA_SCALE: Sym = Variable("GAMMA_SCALE")
    GAMMA_SHAPE2: Sym = Variable("GAMMA_SHAPE2")
    GAMMA_SCALE2: Sym = Variable("GAMMA_SCALE2")
    MIN_MELT_FACTOR: Sym = Variable("MIN_MELT_FACTOR")
    MAX_MELT_FACTOR: Sym = Variable("MAX_MELT_FACTOR")
    DD_MELT_TEMP: Sym = Variable("DD_MELT_TEMP")
    DD_AGGRADATION: Sym = Variable("DD_AGGRADATION")
    SNOW_SWI_MIN: Sym = Variable("SNOW_SWI_MIN")
    SNOW_SWI_MAX: Sym = Variable("SNOW_SWI_MAX")
    SWI_REDUCT_COEFF: Sym = Variable("SWI_REDUCT_COEFF")
    DD_REFREEZE_TEMP: Sym = Variable("DD_REFREEZE_TEMP")
    REFREEZE_FACTOR: Sym = Variable("REFREEZE_FACTOR")
    REFREEZE_EXP: Sym = Variable("REFREEZE_EXP")
    PET_CORRECTION: Sym = Variable("PET_CORRECTION")
    HMETS_RUNOFF_COEFF: Sym = Variable("HMETS_RUNOFF_COEFF")
    PERC_COEFF: Sym = Variable("PERC_COEFF")
    BASEFLOW_COEFF_1: Sym = Variable("BASEFLOW_COEFF_1")
    BASEFLOW_COEFF_2: Sym = Variable("BASEFLOW_COEFF_2")
    TOPSOIL: Sym = Variable("TOPSOIL")
    PHREATIC: Sym = Variable("PHREATIC")


class ForestHRU(HRU):
    land_use_class: str = "FOREST"
    veg_class: str = "FOREST"
    soil_profile: str = "DEFAULT_P"
    aquifer_profile: str = "[NONE]"
    terrain_class: str = "[NONE]"
    hru_type: Literal["land"] = "land"


class HRUs(rc.HRUs):
    """
    HRUs command for HMETS.

    Pydantic is able to automatically detect if an HRU is Land or Lake if `hru_type` is provided.
    """

    root: Sequence[ForestHRU]


class HMETS(Config):
    """
    Hydrology Model - École de technologie supérieure (HMETS).

    References
    ----------
    Martel, J.-L., Demeester, K., Brissette, F., Arsenault, R., Poulin, A. 2017.
    HMETS: A simple and efficient hydrology model for teaching hydrological modelling,
    flow forecasting and climate change impacts. Int. J. Eng. Educ., 33, 1307–1316.
    """

    params: P = P()
    hrus: HRUs = Field([ForestHRU()], alias="HRUs")
    netcdf_attribute: dict[str, str] = {"model_id": "HMETS"}
    sub_basins: rc.SubBasins = Field([rc.SubBasin()], alias="SubBasins")
    write_netcdf_format: bool = Field(True, alias="WriteNetcdfFormat")
    time_step: Union[float, str] = Field(1.0, alias="TimeStep")
    calendar: o.Calendar = Field("PROLEPTIC_GREGORIAN", alias="Calendar")
    uniform_initial_conditions: dict[str, Sym] = Field(
        {"SOIL[0]": P.TOPSOIL / 2, "SOIL[1]": P.PHREATIC / 2},
        alias="UniformInitialConditions",
    )
    potential_melt_method: o.PotentialMeltMethod = Field("POTMELT_HMETS", alias="PotentialMeltMethod")
    rain_snow_fraction: o.RainSnowFraction = Field("RAINSNOW_DATA", alias="RainSnowFraction")
    evaporation: o.Evaporation = Field("PET_OUDIN", alias="Evaporation")
    catchment_route: o.CatchmentRoute = Field("ROUTE_DUMP", alias="CatchmentRoute")
    routing: o.Routing = Field("ROUTE_NONE", alias="Routing")
    soil_model: rc.SoilModel = Field(2, alias="SoilModel")

    hydrologic_processes: Sequence[Process] = Field(
        [
            p.SnowBalance(algo="SNOBAL_HMETS", source="MULTIPLE", to="MULTIPLE"),
            p.Precipitation(algo="RAVEN_DEFAULT", source="ATMOS_PRECIP", to="MULTIPLE"),
            p.Infiltration(algo="INF_HMETS", source="PONDED_WATER", to="MULTIPLE"),
            p.Overflow(algo="OVERFLOW_RAVEN", source="SOIL[0]", to="CONVOLUTION[1]"),
            p.Baseflow(algo="BASE_LINEAR", source="SOIL[0]", to="SURFACE_WATER"),
            p.Percolation(algo="PERC_LINEAR", source="SOIL[0]", to="SOIL[1]"),
            p.Overflow(algo="OVERFLOW_RAVEN", source="SOIL[1]", to="CONVOLUTION[1]"),
            p.SoilEvaporation(algo="SOILEVAP_ALL", source="SOIL[0]", to="ATMOSPHERE"),
            p.Convolve(algo="CONVOL_GAMMA", source="CONVOLUTION[0]", to="SURFACE_WATER"),
            p.Convolve(algo="CONVOL_GAMMA_2", source="CONVOLUTION[1]", to="SURFACE_WATER"),
            p.Baseflow(algo="BASE_LINEAR", source="SOIL[1]", to="SURFACE_WATER"),
        ]
    )

    soil_classes: rc.SoilClasses = Field([{"name": "TOPSOIL"}, {"name": "PHREATIC"}], alias="SoilClasses")
    land_use_classes: rc.LandUseClasses = Field([{"name": "FOREST", "forest_coverage": 1}], alias="LandUseClasses")
    vegetation_classes: rc.VegetationClasses = Field(
        [{"name": "FOREST", "max_ht": 4, "max_lai": 5, "max_leaf_cond": 5}],
        alias="VegetationClasses",
    )
    soil_profiles: rc.SoilProfiles = Field(
        [
            {"name": "LAKE"},
            {"name": "ROCK"},
            {
                "name": "DEFAULT_P",
                "soil_classes": ("TOPSOIL", "PHREATIC"),
                "thicknesses": (P.TOPSOIL / 1000.0, P.PHREATIC / 1000.0),
            },
        ]
    )
    global_parameter: dict[str, Sym] = Field(
        {
            "SNOW_SWI_MIN": P.SNOW_SWI_MIN,
            "SNOW_SWI_MAX": P.SNOW_SWI_MAX,
            "SWI_REDUCT_COEFF": P.SWI_REDUCT_COEFF,
            "SNOW_SWI": 0.05,
        },
        alias="GlobalParameter",
    )
    soil_parameter_list: rc.SoilParameterList = Field(
        {
            "parameters": (
                "POROSITY",
                "PERC_COEFF",
                "PET_CORRECTION",
                "BASEFLOW_COEFF",
            ),
            "pl": [
                PL(
                    name="TOPSOIL",
                    values=(1, P.PERC_COEFF, P.PET_CORRECTION, P.BASEFLOW_COEFF_1),
                ),
                PL(name="PHREATIC", values=(1.0, 0, 0, P.BASEFLOW_COEFF_2)),
            ],
        },
        alias="SoilParameterList",
    )
    land_use_parameter_list: rc.LandUseParameterList = Field(
        {
            "parameters": [
                "MIN_MELT_FACTOR",
                "MAX_MELT_FACTOR",
                "DD_MELT_TEMP",
                "DD_AGGRADATION",
                "REFREEZE_FACTOR",
                "REFREEZE_EXP",
                "DD_REFREEZE_TEMP",
                "HMETS_RUNOFF_COEFF",
                "GAMMA_SHAPE",
                "GAMMA_SCALE",
                "GAMMA_SHAPE2",
                "GAMMA_SCALE2",
            ],
            "pl": [
                PL(
                    name="[DEFAULT]",
                    values=(
                        P.MIN_MELT_FACTOR,
                        P.MAX_MELT_FACTOR,
                        P.DD_MELT_TEMP,
                        P.DD_AGGRADATION,
                        P.REFREEZE_FACTOR,
                        P.REFREEZE_EXP,
                        P.DD_REFREEZE_TEMP,
                        P.HMETS_RUNOFF_COEFF,
                        P.GAMMA_SHAPE,
                        P.GAMMA_SCALE,
                        P.GAMMA_SHAPE2,
                        P.GAMMA_SCALE2,
                    ),
                )
            ],
        },
        alias="LandUseParameterList",
    )
    vegetation_parameter_list: rc.VegetationParameterList = Field(
        {
            "parameters": ["RAIN_ICEPT_PCT", "SNOW_ICEPT_PCT"],
            "pl": [PL(name="[DEFAULT]", values=(0, 0))],
        },
        alias="VegetationParameterList",
    )
    _nc_attrs = field_validator("netcdf_attribute")(nc_attrs)



from ravenpy.config.emulators import BasicRoute
from ravenpy.extractors.routing_product import BasinMakerExtractor, open_shapefile


def test_basinmaker_extractor(tmp_path, yangtze):
    routing_product_shp_path = yangtze.fetch("basinmaker/drainage_region_0175_v2-1/finalcat_info_v2-1.zip")
    df = open_shapefile(
        routing_product_shp_path,
    )
    rvh_extractor = BasinMakerExtractor(
        df=df,
        hru_aspect_convention="ArcGIS",
        routing_product_version="2.1",
    )
    rvh_config = rvh_extractor.extract(hru_from_sb=True)

    # Create lists of values to check
    bedslope_list = [item["bed_slope"] for item in rvh_config["channel_profile"]]
    mannings_list = [value for d in rvh_config["channel_profile"] for value in [t[1] for t in d["roughness_zones"]]]
    reach_length_list = [item["reach_length"] for item in rvh_config["sub_basins"]]

    rvh_config.pop("channel_profile")

    config = BasicRoute(**rvh_config)
    config.write_rv(tmp_path, modelname="routing")

    # Checks that the bedslope, Manning and reach length values are non negative numbers
    assert all(isinstance(x, (int, float)) for x in bedslope_list) is True
    assert any(x < 0 for x in bedslope_list) is False
    assert all(isinstance(y, (int, float)) for y in mannings_list) is True
    assert any(y < 0 for y in mannings_list) is False
    assert all(isinstance(z, (int, float)) for z in bedslope_list) is True
    assert any(z < 0 for z in reach_length_list) is False



#!/usr/bin/env python
#
# ravenpy documentation build configuration file, created by
# sphinx-quickstart on Fri Jun  9 13:47:02 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
#
import os
import sys
from datetime import date
from pathlib import Path
from typing import Any

sys.path.insert(0, os.path.abspath(".."))

import ravenpy  # noqa: E402

# -- Workarounds ------------------------------------------------------


def rebuild_readme():
    """Rebuild the readme.rst file from the top-level README.rst file.

    This is a workaround to remove the badge table found in the README.rst
    when building the docs specifically for latex/pdf output.
    """

    with Path("../README.rst").open(encoding="utf-8") as f:
        readme = f.read()

    # Remove the badge table
    readme = readme.replace(
        "=======\nRavenPy\n=======", "=======\nRavenPy\n=======\n\n.. only:: not latex"
    )
    readme = readme.replace("\n+-", "\n    +-")
    readme = readme.replace("\n|", "\n    |")

    with Path("readme.rst").open("w", encoding="utf-8") as f:
        f.write(readme)


rebuild_readme()


# -- General configuration ---------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "IPython.sphinxext.ipython_console_highlighting",
    "myst_nb",
    "sphinx_click",
    "sphinx_codeautolink",
    "sphinx_copybutton",
    "sphinxcontrib.cairosvgconverter",
    "sphinxcontrib.autodoc_pydantic",
]

linkcheck_ignore = [
    r"https://www.ouranos.ca",  # bad ssl certificate
    # Added on 2023-03-06: Wiley does not allow linkcheck requests (error 403)
    r"https://doi.org/10.1029/2020WR029229",
]

nbsphinx_custom_formats = {
    ".md": ["jupytext.reads", {"fmt": "mystnb"}],
}
autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 2

autosummary_generate = True

nb_execution_mode = "cache"
nb_execution_allow_errors = False
nb_execution_excludepatterns = [
    "configuration.md",
    "notebooks/*.ipynb",
    "notebooks/paper/*.ipynb",
]

# nbsphinx_execute = "auto"
# nbsphinx_timeout = 1
# nbsphinx_allow_errors = True

extlinks = {
    "issue": ("https://github.com/CSHS-CWRA/RavenPy/issues/%s", "GH/%s"),
    "pull": ("https://github.com/CSHS-CWRA/RavenPy/pull/%s", "PR/%s"),
    "user": ("https://github.com/%s", "@%s"),
}

# To ensure that underlined fields (e.g. `_field`) are shown in the docs.
autodoc_default_options = {
    "members": True,
    "private-members": False,
    "special-members": False,
    "undoc-members": True,
}

# To avoid having to install these and burst memory limit on ReadTheDocs.
autodoc_mock_imports = [
    "affine",
    "cftime",
    "cf_xarray",
    "click",
    "climpred",
    "clisops",
    "fiona",
    "gdal",
    "h5netcdf",
    "netCDF4",
    "osgeo",
    "geopandas",
    "haversine",
    "holoviews",
    "hvplot",
    "lxml",
    "owslib",
    "pandas",
    "pyproj",
    "rasterio",
    "rioxarray",
    "scipy",
    "shapely",
    "spotpy",
    "statsmodels",
    "xarray",
    "xclim",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a dictionary of suffix: filetype
source_suffix = {".rst": "restructuredtext"}

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "RavenPy"
copyright = f"2021-{date.today().year}, David Huard"
author = "David Huard"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = ravenpy.__version__.split("-")[0]
# The full version, including alpha/beta/rc tags.
release = ravenpy.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# Sphinx-intl configuration
locale_dirs = ["locales/"]
gettext_compact = False  # optional

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    ".jupyter_cache",
    "jupyter_execute",
    "notebooks/paper",
    "notebooks-notworking",
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# Suppress "WARNING: unknown mimetype for ..." when building EPUB.
suppress_warnings = ["epub.unknown_project_files", "mystnb.unknown_mime_type"]

# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a
# theme further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {"style_external_links": True, "logo_only": True}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
if not os.path.exists("_static"):
    os.makedirs("_static")
html_static_path = ["_static"]

html_logo = "_static/_images/logos/ravenpy-logo.png"

# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "ravenpydoc"

# -- Options for LaTeX output ------------------------------------------

latex_elements: dict[str, Any] = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    "papersize": "letterpaper",
    # The font size ('10pt', '11pt' or '12pt').
    #
    "pointsize": "10pt",
    # Additional stuff for the LaTeX preamble.
    #
    "preamble": "",
    # Latex figure (float) alignment
    #
    "figure_align": "htbp",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (master_doc, "ravenpy.tex", "RavenPy Documentation", "David Huard", "manual"),
]

# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "ravenpy", "RavenPy Documentation", [author], 1)]

# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "ravenpy",
        "RavenPy Documentation",
        author,
        "ravenpy",
        "A Python wrapper to setup and run the hydrologic modelling framework Raven.",
        "Miscellaneous",
    ),
]



from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr


__all__ = [
    "count_pixels",
    "make_bnds",
    "synthetic_gr4j_inputs",
]


def count_pixels(stats: dict, numeric_categories=False) -> int:
    category_counts = 0
    for key, val in stats.items():
        if numeric_categories:
            try:
                int(key)
            except ValueError:  # noqa: S112
                continue
        if key in ["count", "min", "max", "mean", "median", "sum", "nodata"]:
            continue
        category_counts += val
    return category_counts


def make_bnds(params, delta):
    """
    Return low and high parameter bounds by subtracting and adding delta*params to params.

    Parameters
    ----------
    params : sequence
        Parameters.
    delta : float [0,1]
        Relative delta to subtract and add to parameters.

    Returns
    -------
    (tuple, tuple)
        Low and high bounds for parameters.
    """
    arr = np.asarray(params)
    d = np.abs(arr * delta)
    return tuple(arr - d), tuple(arr + d)


def synthetic_gr4j_inputs(path):
    time = pd.date_range(start="2000-07-01", end="2002-07-01", freq="D")

    pr = 3 * np.ones(len(time))
    pr = xr.DataArray(pr, coords={"time": time}, dims="time", name="pr")
    pr.to_netcdf(Path(path).joinpath("pr.nc"))

    tas = 280 + 20 * np.cos(np.arange(len(time)) * 2 * np.pi / 365.0)
    tas = xr.DataArray(tas, coords={"time": time}, dims="time", name="tas")
    tas.to_netcdf(Path(path).joinpath("tas.nc"))

    evap = 3 + 3 * np.cos(-30 + np.arange(len(time)) * 2 * np.pi / 365.0)
    evap = xr.DataArray(evap, coords={"time": time}, dims="time", name="evap")
    evap.to_netcdf(Path(path).joinpath("evap.nc"))



from shutil import copyfile

import numpy as np
from xarray import open_dataset
from xclim import set_options
from xclim.indicators.generic import fit, stats

from ravenpy.utilities import graphs


class TestGraph:
    def test_ts_fit_graph(self, tmp_path, yangtze):
        raven_hydrograph = yangtze.fetch(
            "hydro_simulations/raven-gr4j-cemaneige-sim_hmets-0_Hydrographs.nc",
        )
        file = tmp_path / "raven-gr4j-cemaneige-sim_hmets-0_Hydrographs.nc"

        copyfile(raven_hydrograph, file)

        with open_dataset(file) as ds:
            ts = stats(ds.q_sim, op="max", freq="ME")
            with set_options(check_missing="skip"):
                p = fit(ts)
            np.testing.assert_array_equal(p.isnull(), False)

            fig = graphs.ts_fit_graph(ts, p)
            assert fig



from __future__ import annotations
import logging
import os
from pathlib import Path

import pytest
import xarray as xr

# Additional pytest fixtures for emulators
from emulators import (  # noqa: F401
    config_rv,
    gr4jcn_config,
    minimal_emulator,
    numeric_config,
    symbolic_config,
)
from xclim.indicators.generic import fit, stats

from ravenpy.testing.utils import (
    TESTDATA_BRANCH,
    TESTDATA_CACHE_DIR,
    TESTDATA_REPO_URL,
    default_testdata_cache,
    gather_testing_data,
    testing_setup_warnings,
)
from ravenpy.testing.utils import open_dataset as _open_dataset
from ravenpy.testing.utils import yangtze as _yangtze


def convert_2d(fn):
    """
    Take the 1D Salmon time series and convert it to a 2D time series.

    Example
    -------
    >>> fn = "./testdata/raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc"
    >>> fn2 = "./testdata/raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily_2d.nc"
    >>> _convert_2d(fn).to_netcdf(fn2, "w")
    """
    features = {
        "name": "Salmon",
        "area": 4250.6,
        "elevation": 843.0,
        "latitude": 54.4848,
        "longitude": -123.3659,
    }
    ds = xr.open_dataset(fn, decode_times=False).rename({"nstations": "region"})

    out = xr.Dataset(
        coords={
            "lon": ds.lon.expand_dims("lon").squeeze("region"),
            "lat": ds.lat.expand_dims("lat").squeeze("region"),
            "time": ds.time,
        }
    )

    for v in ds.data_vars:
        if v not in ["lon", "lat"]:
            out[v] = ds[v].expand_dims("region", axis=1)

    # Add geometry feature variables
    for key, val in features.items():
        out[key] = xr.DataArray(name=key, data=[val], dims="region")

    return out


def convert_3d(fn):
    """
    Take the 1D Salmon time series and convert it to a 3D time series.

    Example
    -------
    >>> fn = "./testdata/raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc"
    >>> fn3 = "./testdata/raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily_3d.nc"
    >>> _convert_3d(fn).to_netcdf(fn3, "w")
    """
    elevation = [[843.0]]
    ds = xr.open_dataset(fn, decode_times=False)

    out = xr.Dataset(
        coords={
            "lon": ds.lon.expand_dims("lon").squeeze("nstations"),
            "lat": ds.lat.expand_dims("lat").squeeze("nstations"),
            "time": ds.time,
        }
    )

    for v in ds.data_vars:
        if v not in ["lon", "lat", "time"]:
            out[v] = ds[v]
            out[v] = out[v].expand_dims(["lon", "lat"])  # Needs to be in other step to keep attributes

    out["elevation"] = xr.DataArray(
        data=elevation,
        dims=["lon", "lat"],
        attrs={"units": "m", "standard_name": "altitude"},
    )

    return out


@pytest.fixture(scope="session")
def threadsafe_data_dir(tmp_path_factory) -> Path:
    return Path(tmp_path_factory.getbasetemp().joinpath("data"))


@pytest.fixture(scope="session")
def yangtze(threadsafe_data_dir, worker_id):
    return _yangtze(
        repo=TESTDATA_REPO_URL,
        branch=TESTDATA_BRANCH,
        cache_dir=(TESTDATA_CACHE_DIR if worker_id == "master" else threadsafe_data_dir),
    )


@pytest.fixture(scope="session")
def open_dataset(threadsafe_data_dir, worker_id):
    def _open_session_scoped_file(file: str | os.PathLike, **xr_kwargs):
        _yangtze_kwargs = {
            "branch": TESTDATA_BRANCH,
            "repo": TESTDATA_REPO_URL,
            "cache_dir": (TESTDATA_CACHE_DIR if worker_id == "master" else threadsafe_data_dir),
        }
        xr_kwargs.setdefault("cache", True)
        xr_kwargs.setdefault("engine", "h5netcdf")
        return _open_dataset(
            file,
            _yangtze_kwargs=_yangtze_kwargs,
            **xr_kwargs,
        )

    return _open_session_scoped_file


@pytest.fixture(scope="session")
def q_sim_1(yangtze):
    """A file storing a Raven streamflow simulation over one basin."""
    return yangtze.fetch(
        "hydro_simulations/raven-gr4j-cemaneige-sim_hmets-0_Hydrographs.nc",
    )


@pytest.fixture(scope="session")
def ts_stats(q_sim_1, threadsafe_data_dir):
    fn = threadsafe_data_dir / "ts_stats.nc"

    if not fn.exists():
        with xr.open_dataset(q_sim_1) as ds:
            q = ds.q_sim
            ts = stats(q, op="max")
            ts.to_netcdf_(fn)
    return fn


@pytest.fixture(scope="session")
def fit_params(ts_stats, threadsafe_data_dir):
    fn = threadsafe_data_dir / "fit.nc"

    if not fn.exists():
        with xr.open_dataset(ts_stats) as ds:
            name = list(ds.data_vars.keys()).pop()
            q = ds[name]
            p = fit(q, dist="gumbel_r")
            p.to_netcdf(fn)
    return fn


@pytest.fixture
def bad_netcdf(yangtze, tmp_path):
    fn = tmp_path / "bad_netcdf.nc"

    salmon_file = yangtze.fetch("raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc")

    # Test with scalar elevation. Should normally have a station dimension, but not always the case.
    with xr.open_dataset(salmon_file) as ds:
        ds["station_name"] = ds["station_name"].astype("str")
        ds["elevation"] = 1.0
        ds.to_netcdf(fn)
    return fn


@pytest.fixture(scope="session")
def salmon_hru():
    out = {
        "land": {
            "area": 4250.6,
            "elevation": 843.0,
            "latitude": 54.4848,
            "longitude": -123.3659,
            "hru_type": "land",
        }
    }
    return out


# Used in test_emulators.py
@pytest.fixture(scope="session")
def input2d(salmon, threadsafe_data_dir):
    """Convert 1D input to 2D output by copying all the time series along a new region dimension."""
    fn_out = threadsafe_data_dir / "input2d.nc"

    if not fn_out.exists():
        convert_2d(salmon).to_netcdf(fn_out)

    return fn_out


@pytest.fixture(scope="session")
def input3d(salmon, threadsafe_data_dir):
    """Convert 1D input to 2D output by copying all the time series along a new region dimension."""
    fn_out = threadsafe_data_dir / "input3d.nc"

    if not fn_out.exists():
        ds = convert_3d(salmon)
        ds = ds.drop_vars("qobs")
        ds.to_netcdf(fn_out)
        ds.close()

    return fn_out


@pytest.fixture
def dummy_config():
    """Return an almost empty config class and the parameter dataclass."""
    from pydantic import Field
    from pydantic.dataclasses import dataclass, rebuild_dataclass

    from ravenpy.config import options as o
    from ravenpy.config.base import Params, Sym, SymConfig, Variable
    from ravenpy.config.rvs import Config

    @dataclass(config=SymConfig)
    class P(Params):
        X1: Sym = Variable("X1")

    rebuild_dataclass(P)

    class TestConfig(Config):
        params: P = P()
        calendar: o.Calendar = Field("JULIAN", alias="Calendar")
        air_snow_coeff: Sym = Field(1 - P.X1, alias="AirSnowCoeff")

    return TestConfig, P


@pytest.fixture(autouse=True, scope="session")
def gather_session_data(request, yangtze, worker_id):
    """
    Gather testing data on pytest run.

    When running pytest with multiple workers, one worker will copy data remotely to the default cache dir while
    other workers wait using a lockfile. Once the lock is released, all workers will then copy data to their local
    threadsafe_data_dir. As this fixture is scoped to the session, it will only run once per pytest run.
    """
    testing_setup_warnings()
    gather_testing_data(worker_cache_dir=yangtze.path, worker_id=worker_id)

    def remove_data_written_flag():
        """Clean up the cache folder once we are finished."""
        flag = default_testdata_cache.joinpath(".data_written")
        if flag.exists():
            try:
                flag.unlink()
            except FileNotFoundError:
                logging.info("Teardown race condition occurred: .data_written flag already removed. Lucky!")
                pass

    request.addfinalizer(remove_data_written_flag)



import pathlib
import re
from collections.abc import Sequence
from textwrap import dedent
from typing import Union

import pytest
import xarray as xr
from pydantic import Field

from ravenpy.config import commands as rc
from ravenpy.config import options as o
from ravenpy.config.base import RV, optfield


salmon_land_hru_1 = dict(area=4250.6, elevation=843.0, latitude=54.4848, longitude=-123.3659, hru_type="land")


def test_custom_output():
    class Test(RV):
        co: Sequence[rc.CustomOutput] = optfield(alias="CustomOutput")

    co = rc.CustomOutput(
        time_per="YEARLY",
        stat="AVERAGE",
        variable="PRECIP",
        space_agg="ENTIRE_WATERSHED",
    )
    assert Test(co=[co]).to_rv().strip() == ":CustomOutput         YEARLY AVERAGE PRECIP ENTIRE_WATERSHED"

    new = rc.CustomOutput.parse(str(co))
    assert new == co


def test_evaluation_metrics():
    evaluation_metrics = [
        "ABSERR",
        "ABSERR_RUN",
        "ABSMAX",
        "DIAG_SPEARMAN",
        "FUZZY_NASH",
        "KGE_PRIME",
        "KLING_GUPTA",
        "LOG_NASH",
        "NASH_SUTCLIFFE",
        "NSC",
        "PCT_BIAS",
        "PDIFF",
        "RCOEFF",
        "RMSE",
        "TMVOL",
    ]

    class Test(RV):
        em: Sequence[o.EvaluationMetrics] = Field(evaluation_metrics, alias="EvaluationMetrics")

    assert Test().to_rv().strip() == f":EvaluationMetrics    {' '.join(evaluation_metrics)}"


def test_hrus():
    hrus = rc.HRUs([salmon_land_hru_1])
    hrus.to_rv()
    assert hrus[0].subbasin_id == 1


def test_hrus_parse(yangtze):
    f = yangtze.fetch("clrh/mattawin/06FB002.rvh")
    hrus = rc.HRUs.parse(pathlib.Path(f).read_text())
    assert len(hrus) == 40
    hru = hrus[0]

    assert hru.hru_id == 38
    assert hru.area == 2.1924
    assert hru.elevation == 1440.5898
    assert hru.latitude == 46.6007
    assert hru.longitude == -73.9222
    assert hru.subbasin_id == 23007946
    assert hru.land_use_class == "LAKE"
    assert hru.veg_class == "LAKE"
    assert hru.soil_profile == "LAKE"
    assert hru.aquifer_profile == "[NONE]"
    assert hru.terrain_class == "[NONE]"
    assert hru.slope == 30.986380
    assert hru.aspect == 190.30589321792877


def test_hru_state():
    s = rc.HRUState(hru_id=1, data={"SOIL[0]": 1, "SOIL[1]": 2.0})
    assert str(s) == "1,1.0,2.0"


def test_linear_transform():
    assert rc.LinearTransform(scale=1, offset=0).to_rv() == ""


def test_global_parameter():
    class Test(RV):
        gp: dict = Field({"A": 1}, alias="GlobalParameter")

    s = Test().to_rv().strip()

    assert s == ":GlobalParameter A 1"


def test_vegetation_classes():
    class Test(RV):
        vegetation_classes: rc.VegetationClasses = Field(
            rc.VegetationClasses([{"name": "VEG_ALL"}, {"name": "VEG_WATER"}]),
            alias="VegetationClasses",
        )

    s = dedent(Test().to_rv().strip())
    assert (
        s
        == dedent(
            """
    :VegetationClasses
      :Attributes     ,            MAX_HT,           MAX_LAI,     MAX_LEAF_COND
      :Units          ,                 m,              none,          mm_per_s
      VEG_ALL         ,               0.0,               0.0,               0.0
      VEG_WATER       ,               0.0,               0.0,               0.0
    :EndVegetationClasses"""
        ).strip()
    )


def test_land_use_class():
    class Test(RV):
        lu_classes: rc.LandUseClasses = Field(None, alias="LandUseClasses")

    t = Test(
        LandUseClasses=[
            {"name": "Lake"},
            {"name": "Land", "impermeable_frac": 0.1, "forest_coverage": 0.8},
        ]
    )
    s = dedent(t.to_rv()).strip()
    assert (
        s
        == dedent(
            """
        :LandUseClasses
          :Attributes     ,  IMPERMEABLE_FRAC,   FOREST_COVERAGE
          :Units          ,             fract,             fract
          Lake            ,               0.0,               0.0
          Land            ,               0.1,               0.8
        :EndLandUseClasses
        """
        ).strip()
    )


def test_soil_classes():
    from ravenpy.config.commands import SoilClasses

    c = SoilClasses([{"name": "TOPSOIL"}, {"name": "FAST_RES"}, {"name": "SLOW_RES"}])
    assert dedent(c.to_rv()) == dedent(
        """
    :SoilClasses
      :Attributes     ,             %SAND,             %CLAY,             %SILT,          %ORGANIC
      :Units          ,              none,              none,              none,              none
      TOPSOIL
      FAST_RES
      SLOW_RES
    :EndSoilClasses
    """
    )

    c = SoilClasses(
        [
            {"name": "TOPSOIL", "mineral": (0.1, 0.7, 0.2), "organic": 0.1},
        ]
    )
    assert dedent(c.to_rv()) == dedent(
        """
    :SoilClasses
      :Attributes     ,             %SAND,             %CLAY,             %SILT,          %ORGANIC
      :Units          ,              none,              none,              none,              none
      TOPSOIL         ,               0.1,               0.7,               0.2,               0.1
    :EndSoilClasses
    """
    )


def test_soil_model():
    class Test(RV):
        soil_model: rc.SoilModel = Field(None, alias="SoilModel")

    t = Test(soil_model=3)
    assert t.to_rv().strip() == ":SoilModel            SOIL_MULTILAYER 3"


def test_soil_profiles(x=42.0):
    c = rc.SoilProfiles(
        [
            {
                "name": "DEFAULT_P",
                "soil_classes": ["TOPSOIL", "FAST_RES", "SLOW_RES"],
                "thicknesses": [x, 100.0, 100.0],
            },
        ]
    )

    pat = r"DEFAULT_P\s*,\s*3,\s*TOPSOIL,\s*(.+),\s*FAST_RES,\s*100.0,\s*SLOW_RES,\s*100.0"

    m = re.findall(pat, str(c))[0]
    assert m == "42.0"


def test_hydrologic_processes():
    from ravenpy.config.processes import Precipitation, SnowTempEvolve

    hp = [
        Precipitation(algo="PRECIP_RAVEN", source="ATMOS_PRECIP", to="MULTIPLE"),
        SnowTempEvolve(algo="SNOTEMP_NEWTONS", source="SNOW_TEMP"),
    ]

    class Test(RV):
        hydrologic_processes: Sequence[rc.Process] = Field(hp, alias="HydrologicProcesses")

    out = dedent(Test().to_rv().strip())
    assert (
        out
        == dedent(
            """
        :HydrologicProcesses
          :Precipitation        PRECIP_RAVEN        ATMOS_PRECIP        MULTIPLE
          :SnowTempEvolve       SNOTEMP_NEWTONS     SNOW_TEMP
        :EndHydrologicProcesses
        """
        ).strip()
    )


def test_process_group():
    from ravenpy.config.processes import Precipitation, ProcessGroup, SnowTempEvolve

    hp = [
        Precipitation(algo="PRECIP_RAVEN", source="ATMOS_PRECIP", to="MULTIPLE"),
        SnowTempEvolve(algo="SNOTEMP_NEWTONS", source="SNOW_TEMP"),
    ]
    c = ProcessGroup(p=hp, params=(0, 1))
    out = dedent(c.to_rv().strip())

    assert (
        out
        == dedent(
            """
    :ProcessGroup
        :Precipitation        PRECIP_RAVEN        ATMOS_PRECIP        MULTIPLE
        :SnowTempEvolve       SNOTEMP_NEWTONS     SNOW_TEMP
    :EndProcessGroup CALCULATE_WTS 0.0 1.0
        """
        ).strip()
    )


def test_soil_parameter_list():
    c = rc.SoilParameterList(
        parameters=["POROSITY", "FIELD_CAPACITY"],
        pl=[
            rc.PL(name="[DEFAULT]", values=[1, 0]),
            rc.PL(name="FAST_RES", values=[None, None]),
        ],
    )

    assert dedent(c.to_rv()) == dedent(
        """
    :SoilParameterList
      :Parameters     ,          POROSITY,    FIELD_CAPACITY
      :Units          ,              none,              none
      [DEFAULT]       ,               1.0,               0.0
      FAST_RES        ,          _DEFAULT,          _DEFAULT
    :EndSoilParameterList
    """
    )


def test_vegetation_parameter_list():
    c = rc.VegetationParameterList(
        parameters=[
            "MAX_CAPACITY",
            "MAX_SNOW_CAPACITY",
            "RAIN_ICEPT_PCT",
            "SNOW_ICEPT_PCT",
        ],
        pl=[rc.PL(name="VEG_ALL", values=[10000, 10000, 0.88, 0.88])],
    )

    assert dedent(c.to_rv()) == dedent(
        """
    :VegetationParameterList
      :Parameters     ,      MAX_CAPACITY, MAX_SNOW_CAPACITY,    RAIN_ICEPT_PCT,    SNOW_ICEPT_PCT
      :Units          ,              none,              none,              none,              none
      VEG_ALL         ,           10000.0,           10000.0,              0.88,              0.88
    :EndVegetationParameterList
    """
    )


def test_land_use_parameter_list():
    c = rc.LandUseParameterList(
        parameters=["MELT_FACTOR", "MIN_MELT_FACTOR"],
        pl=[rc.PL(name="[DEFAULT]", values=[1, 2.2])],
    )
    assert dedent(c.to_rv()) == dedent(
        """
    :LandUseParameterList
      :Parameters     ,       MELT_FACTOR,   MIN_MELT_FACTOR
      :Units          ,              none,              none
      [DEFAULT]       ,               1.0,               2.2
    :EndLandUseParameterList
    """
    )

    with pytest.raises(ValueError):
        rc.LandUseParameterList(
            names=["MELT_FACTOR", "MIN_MELT_FACTOR"],
            pl=[rc.PL(name="[DEFAULT]", values=[None, 2.2])],
        )


class TestHRUStateVariableTable:
    def test_to_rv(self):
        s1 = rc.HRUState(hru_id=1, data={"SOIL[0]": 0.1, "SOIL[1]": 1.0})
        s2 = rc.HRUState(hru_id=2, data={"SOIL[3]": 3, "SOIL[2]": 2.0})
        t = rc.HRUStateVariableTable([s1, s2])
        assert dedent(t.to_rv()) == dedent(
            """
            :HRUStateVariableTable
              :Attributes     ,           SOIL[0],           SOIL[1],           SOIL[2],           SOIL[3]
              1,0.1,1.0,0.0,0.0
              2,0.0,0.0,2.0,3.0
            :EndHRUStateVariableTable
            """
        )

        s1 = dict(hru_id=1, data={"SOIL[0]": 0.1, "SOIL[1]": 1.0})
        rc.HRUStateVariableTable([s1])

    def test_parse(self):
        solution = """
:HRUStateVariableTable
  :Attributes,SURFACE_WATER,ATMOSPHERE,ATMOS_PRECIP,PONDED_WATER,SOIL[0],SOIL[1],SNOW,SNOW_LIQ,CUM_SNOWMELT,CONVOLUTION[0],CONVOLUTION[1],AET,CONV_STOR[0],CONV_STOR[1],CONV_STOR[2],CONV_STOR[3],CONV_STOR[4],CONV_STOR[5],CONV_STOR[6],CONV_STOR[7],CONV_STOR[8],CONV_STOR[9],CONV_STOR[10],CONV_STOR[11],CONV_STOR[12],CONV_STOR[13],CONV_STOR[14],CONV_STOR[15],CONV_STOR[16],CONV_STOR[17],CONV_STOR[18],CONV_STOR[19],CONV_STOR[20],CONV_STOR[21],CONV_STOR[22],CONV_STOR[23],CONV_STOR[24],CONV_STOR[25],CONV_STOR[26],CONV_STOR[27],CONV_STOR[28],CONV_STOR[29],CONV_STOR[30],CONV_STOR[31],CONV_STOR[32],CONV_STOR[33],CONV_STOR[34],CONV_STOR[35],CONV_STOR[36],CONV_STOR[37],CONV_STOR[38],CONV_STOR[39],CONV_STOR[40],CONV_STOR[41],CONV_STOR[42],CONV_STOR[43],CONV_STOR[44],CONV_STOR[45],CONV_STOR[46],CONV_STOR[47],CONV_STOR[48],CONV_STOR[49],CONV_STOR[50],CONV_STOR[51],CONV_STOR[52],CONV_STOR[53],CONV_STOR[54],CONV_STOR[55],CONV_STOR[56],CONV_STOR[57],CONV_STOR[58],CONV_STOR[59],CONV_STOR[60],CONV_STOR[61],CONV_STOR[62],CONV_STOR[63],CONV_STOR[64],CONV_STOR[65],CONV_STOR[66],CONV_STOR[67],CONV_STOR[68],CONV_STOR[69],CONV_STOR[70],CONV_STOR[71],CONV_STOR[72],CONV_STOR[73],CONV_STOR[74],CONV_STOR[75],CONV_STOR[76],CONV_STOR[77],CONV_STOR[78],CONV_STOR[79],CONV_STOR[80],CONV_STOR[81],CONV_STOR[82],CONV_STOR[83],CONV_STOR[84],CONV_STOR[85],CONV_STOR[86],CONV_STOR[87],CONV_STOR[88],CONV_STOR[89],CONV_STOR[90],CONV_STOR[91],CONV_STOR[92],CONV_STOR[93],CONV_STOR[94],CONV_STOR[95],CONV_STOR[96],CONV_STOR[97],CONV_STOR[98],CONV_STOR[99]
  :Units,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm,mm
  1,0.00000,0.00000,-0.16005,0.00000,144.54883,455.15708,0.16005,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000,0.00000
:EndHRUStateVariableTable
        """
        sv = rc.HRUStateVariableTable.parse(solution).root
        assert len(sv) == 1
        assert sv[0].hru_id == 1
        assert sv[0].data["ATMOS_PRECIP"] == -0.16005


def test_read_from_netcdf(yangtze):
    from ravenpy.config.commands import ReadFromNetCDF

    f = yangtze.fetch("raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc")
    c = ReadFromNetCDF.from_nc(f, "PRECIP", station_idx=1, alt_names=("rain",))
    s = dedent(c.to_rv())

    pat = re.compile(
        r"""
    :ReadFromNetCDF\s+
      :FileNameNC\s+.+/raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc\s+
      :VarNameNC\s+rain\s+
      :DimNamesNC\s+time\s+
      :StationIdx\s+1\s+
      :LatitudeVarNameNC\s+lat\s+
      :LongitudeVarNameNC\s+lon\s+
    :EndReadFromNetCDF
    """,
        re.VERBOSE + re.MULTILINE,
    )
    assert pat.search(s) is not None
    assert isinstance(c.da, xr.DataArray)


def test_station_forcing(yangtze):
    f = yangtze.fetch("raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily_2d.nc")
    c = rc.StationForcing.from_nc(f, "PRECIP", station_idx=1, alt_names="rain")
    dedent(c.to_rv())


def test_gridded_forcing(yangtze):
    # TODO: Make sure dimensions are in the order x, y, t.
    fn = yangtze.fetch("raven-routing-sample/VIC_temperatures.nc")
    rc.GriddedForcing.from_nc(fn, data_type="TEMP_AVE", alt_names=("Avg_temp",))
    # assert gf.dim_names_nc == ("lon_dim", "lat_dim", "time")

    fn = yangtze.fetch("raven-routing-sample/VIC_streaminputs.nc")
    rc.GriddedForcing.from_nc(fn, data_type="PRECIP", alt_names=("Streaminputs",))
    # assert gf.dim_names_nc == ("lon_dim", "lat_dim", "time")

    fn = yangtze.fetch("cmip5/tas_Amon_CanESM2_rcp85_r1i1p1_200601-210012_subset.nc")
    gf = rc.GriddedForcing.from_nc(fn, data_type="TEMP_AVE", engine="netcdf4")
    assert gf.latitude_var_name_nc is None


def test_gauge(yangtze, tmp_path):
    salmon_file = yangtze.fetch("raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc")
    salmon_file_tmp = tmp_path / "salmon_river_near_prince_george-tmp.nc"
    salmon_file_tmp.write_bytes(pathlib.Path(salmon_file).read_bytes())

    g = rc.Gauge.from_nc(
        salmon_file_tmp,
        alt_names={"RAINFALL": "rain", "SNOWFALL": "snow"},
        data_kwds={"ALL": {"Deaccumulate": True}},
    )

    assert "Data" in g.to_rv()
    assert isinstance(g.ds, xr.Dataset)
    assert "RAINFALL" in g.ds
    assert g.data[0].read_from_netcdf.deaccumulate


def test_gauge_raises(yangtze):
    f = yangtze.fetch("raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc")

    with pytest.raises(ValueError):
        rc.Gauge.from_nc(
            f,
            data_type=[
                "RAINFALL",
            ],
            alt_names={"RAINFALL": "bad_name"},
        )


def test_grid_weights():
    gw = rc.GridWeights()
    txt = gw.to_rv()
    assert len(txt.splitlines()) == 6

    parsed = rc.GridWeights.parse(txt)
    assert parsed.number_hrus == gw.number_hrus
    assert parsed.number_grid_cells == gw.number_grid_cells
    assert parsed.data == gw.data


def test_redirect_to_file(yangtze):
    f = yangtze.fetch("raven-routing-sample/VIC_test_nodata_weights.rvt")
    r = rc.RedirectToFile(f)
    assert re.match(r"^:RedirectToFile\s+(\S+)$", r.to_rv())

    class Test(RV):
        gw: Union[rc.GridWeights, rc.RedirectToFile] = Field(None, alias="RedirectToFile")

    t = Test(gw=f)
    assert t.gw.root == pathlib.Path(f)


def test_subbasin_properties():
    c = rc.SubBasinProperties(
        parameters=["GAMMA_SCALE", "GAMMA_SHAPE"],
        records=[{"sb_id": 1, "values": (0, 1)}],
    )
    assert "0, 1" in c.to_rv()


def test_subbasins():
    c = rc.SubBasins([{"name": "SB1"}])
    c.to_rv()
    assert c[0].gauged

    class Test(RV):
        sub_basins: rc.SubBasins = Field([rc.SubBasin()], alias="SubBasins")
        hrus: rc.HRUs = Field([rc.HRU()], alias="HRUs")

    Test().to_rv()


def test_subbasins_parse(yangtze):
    f = yangtze.fetch("clrh/mattawin/06FB002.rvh")
    sb = rc.SubBasins.parse(pathlib.Path(f).read_text())
    assert len(sb) == 35


def test_reservoir_parse(yangtze):
    f = yangtze.fetch("clrh/mattawin/Lakes.rvh")
    rs = rc.Reservoir.parse(pathlib.Path(f).read_text())
    assert len(rs) == 5
    r = rs[0]

    assert r.name == "Lake_107056"
    assert r.subbasin_id == 23007946
    assert r.hru_id == 38
    assert r.type == "RESROUTE_STANDARD"
    assert r.weir_coefficient == 0.6
    assert r.crest_width == 7.7399
    assert r.max_depth == 9.6
    assert r.lake_area == 2192395.4127609455
    assert r.seepage_parameters.k_seep == 0
    assert r.seepage_parameters.h_ref == 0


def test_ensemble_mode():
    c = rc.EnsembleMode(n=10)
    s = c.to_rv().strip()
    assert s == ":EnsembleMode         ENSEMBLE_ENKF 10"


def test_observation_error_model():
    c = rc.ObservationErrorModel(state="STREAMFLOW", dist="DIST_UNIFORM", p1=1, p2=2, adj="ADDITIVE")

    assert c.to_rv().strip() == ":ObservationErrorModel STREAMFLOW DIST_UNIFORM 1.0 2.0 ADDITIVE"


def test_forcing_perturbation():
    c = rc.ForcingPerturbation(forcing="TEMP_AVE", dist="DIST_UNIFORM", p1=1, p2=2, adj="ADDITIVE")

    assert c.to_rv().strip() == ":ForcingPerturbation  TEMP_AVE DIST_UNIFORM 1.0 2.0 ADDITIVE"


def test_hru_group():
    c = rc.HRUGroup(name="ForestedHRUs", groups=["1", "3-5"])
    s = dedent(c.to_rv()).strip()
    assert (
        s
        == dedent(
            """
            :HRUGroup ForestedHRUs
              1,3-5
            :EndHRUGroup
                """
        ).strip()
    )


def test_subbasin_group_parse(yangtze):
    f = yangtze.fetch("clrh/mattawin/06FB002.rvh")
    sbgs = rc.SubBasinGroup.parse(pathlib.Path(f).read_text())
    assert len(sbgs) == 2
    sbg = sbgs[0]
    assert sbg.name == "Allsubbasins"
    assert len(sbg.sb_ids) == 35
    assert sbg.sb_ids[0] == 23007946


def test_channel_profile_parse(yangtze):
    f = yangtze.fetch("clrh/mattawin/channel_properties.rvp")
    cps = rc.ChannelProfile.parse(pathlib.Path(f).read_text())
    assert len(cps) == 20

    cp = cps[0]

    assert cp.name == "Chn_ZERO_LENGTH"
    assert cp.bed_slope == 0.1234500000
    assert len(cp.survey_points) == 8
    assert cp.survey_points[0].root == (0, 1444.5898)
    assert cp.survey_points[1].root == (16.0000, 1440.5898)
    assert len(cp.roughness_zones) == 3
    assert cp.roughness_zones[0].root == (0, 0.12345000)

    assert str(cp) == dedent(
        """
    :ChannelProfile Chn_ZERO_LENGTH
      :Bedslope             0.12345

      :SurveyPoints
        0.0 1444.5898
        16.0 1440.5898
        16.2469 1440.5898
        16.2778 1440.4663
        16.3395 1440.4663
        16.3703 1440.5898
        16.6172 1440.5898
        32.6173 1444.5898
      :EndSurveyPoints

      :RoughnessZones
        0.0 0.12345
        16.2469 0.12345
        16.37035 0.12345
      :EndRoughnessZones
    :EndChannelProfile
    """
    )



import datetime as dt
import zipfile
from collections.abc import Sequence
from dataclasses import asdict, fields, is_dataclass
from pathlib import Path
from textwrap import dedent
from typing import Any, Optional, Union

import cftime
from pydantic import ConfigDict, Field, ValidationInfo, field_validator

from ..config import commands as rc
from ..config import options as o
from ..config import processes as rp
from .base import RV, Sym, optfield, parse_symbolic


try:
    from raven_hydro import __raven_version__
except ImportError:
    __raven_version__ = "0.0.0"


"""
Generic Raven model configuration.

Note that aliases are set to identify class attributes as Raven commands.
"""
date = Union[dt.date, dt.datetime, cftime.datetime]


class RVI(RV):
    __rv__ = "RVI"

    # Run parameters
    silent_mode: Optional[bool] = optfield(alias="SilentMode")
    noisy_mode: Optional[bool] = optfield(alias="NoisyMode")

    run_name: Optional[str] = optfield(alias="RunName")
    calendar: Optional[o.Calendar] = optfield(alias="Calendar")
    start_date: Optional[Union[str, date]] = optfield(alias="StartDate")
    assimilation_start_time: Optional[date] = optfield(alias="AssimilationStartTime")
    end_date: Optional[Union[str, date]] = optfield(alias="EndDate")
    duration: Optional[float] = optfield(alias="Duration")
    time_step: Optional[Union[float, str]] = optfield(alias="TimeStep")
    interpolation: Optional[o.Interpolation] = optfield(alias="Interpolation")

    # Model description
    routing: Optional[o.Routing] = optfield(alias="Routing")
    catchment_route: Optional[o.CatchmentRoute] = optfield(alias="CatchmentRoute")
    evaporation: Optional[o.Evaporation] = optfield(alias="Evaporation")
    ow_evaporation: Optional[o.Evaporation] = optfield(alias="OW_Evaporation")
    sw_radiation_method: Optional[o.SWRadiationMethod] = optfield(alias="SWRadiationMethod")
    sw_cloud_correct: Optional[o.SWCloudCorrect] = optfield(alias="SWCloudCorrect")
    sw_canopy_correct: Optional[o.SWCanopyCorrect] = optfield(alias="SWCanopyCorrect")
    lw_radiation_method: Optional[o.LWRadiationMethod] = optfield(alias="LWRadiationMethod")
    windspeed_method: Optional[o.WindspeedMethod] = optfield(alias="WindspeedMethod")
    rain_snow_fraction: Optional[o.RainSnowFraction] = optfield(alias="RainSnowFraction")
    potential_melt_method: Optional[o.PotentialMeltMethod] = optfield(alias="PotentialMeltMethod")
    oro_temp_correct: Optional[o.OroTempCorrect] = optfield(alias="OroTempCorrect")
    oro_precip_correct: Optional[o.OroPrecipCorrect] = optfield(alias="OroPrecipCorrect")
    oro_pet_correct: Optional[o.OroPETCorrect] = optfield(alias="OroPETCorrect")
    cloud_cover_method: Optional[o.CloudCoverMethod] = optfield(alias="CloudCoverMethod")
    precip_icept_frac: Optional[o.PrecipIceptFract] = optfield(alias="PrecipIceptFract")
    subdaily_method: Optional[o.SubdailyMethod] = optfield(alias="SubdailyMethod")
    monthly_interpolation_method: Optional[o.MonthlyInterpolationMethod] = optfield(alias="MonthlyInterpolationMethod")
    soil_model: Optional[rc.SoilModel] = optfield(alias="SoilModel")
    temperature_correction: Optional[bool] = optfield(
        alias="TemperatureCorrection",
        description="Gridded or gauged temperature bias correction.",
    )
    lake_storage: Optional[o.StateVariables] = optfield(alias="LakeStorage")
    relative_humidity_method: Optional[o.RelativeHumidityMethod] = optfield(alias="RelativeHumidityMethod")

    define_hru_groups: Optional[Sequence[str]] = optfield(alias="DefineHRUGroups")

    hydrologic_processes: Optional[Sequence[Union[rc.Process, rp.Conditional, rp.ProcessGroup]]] = optfield(alias="HydrologicProcesses")
    evaluation_metrics: Optional[Sequence[o.EvaluationMetrics]] = optfield(alias="EvaluationMetrics")
    evaluation_period: Optional[Sequence[rc.EvaluationPeriod]] = optfield(alias="EvaluationPeriod")
    ensemble_mode: Optional[rc.EnsembleMode] = optfield(alias="EnsembleMode")

    # Options
    write_netcdf_format: Optional[bool] = optfield(alias="WriteNetcdfFormat")
    netcdf_attribute: Optional[dict[str, str]] = optfield(alias="NetCDFAttribute")

    custom_output: Optional[Sequence[rc.CustomOutput]] = optfield(alias="CustomOutput")
    direct_evaporation: Optional[bool] = optfield(
        alias="DirectEvaporation",
        description="Rainfall is automatically reduced through evapotranspiration up to the limit of the calculated PET.",
    )
    deltares_fews_mode: Optional[bool] = optfield(alias="DeltaresFEWSMode")
    debug_mode: Optional[bool] = optfield(alias="DebugMode")
    dont_write_watershed_storage: Optional[bool] = optfield(
        alias="DontWriteWatershedStorage",
        description="Do not write watershed storage variables to disk.",
    )
    pavics_mode: Optional[bool] = optfield(alias="PavicsMode")

    suppress_output: Optional[bool] = optfield(alias="SuppressOutput", description="Write minimal output to disk when enabled.")
    write_forcing_functions: Optional[bool] = optfield(
        alias="WriteForcingFunctions",
        description="Write watershed averaged forcing functions (e.g. rainfall, radiation, PET, etc).",
    )
    write_subbasin_file: Optional[bool] = optfield(alias="WriteSubbasinFile")  # Undocumented
    write_local_flows: Optional[bool] = optfield(
        alias="WriteLocalFlows",
        description="Write local contribution to hydrograph in hydrograph.csv",
    )

    @field_validator("soil_model", mode="before")
    @classmethod
    def init_soil_model(cls, v):
        if isinstance(v, int):
            return rc.SoilModel(v)
        return v

    @field_validator("start_date", "end_date", "assimilation_start_time")
    @classmethod
    def dates2cf(cls, v, info):
        """Convert dates to cftime dates."""
        if v is not None:
            calendar = (info.data.get("calendar") or o.Calendar.PROLEPTIC_GREGORIAN).value.lower()

            obj = cftime._cftime.DATE_TYPES[calendar]
            if isinstance(v, str):
                v = dt.datetime.fromisoformat(v)

            return obj(*v.timetuple()[:6])

    model_config = ConfigDict(arbitrary_types_allowed=True)


class RVT(RV):
    __rv__ = "RVT"

    gauge: Optional[Sequence[rc.Gauge]] = optfield(alias="Gauge")
    station_forcing: Optional[Sequence[rc.StationForcing]] = optfield(alias="StationForcing")
    gridded_forcing: Optional[Sequence[rc.GriddedForcing]] = optfield(alias="GriddedForcing")
    observation_data: Optional[Sequence[rc.ObservationData]] = optfield(alias="ObservationData")


class RVP(RV):
    __rv__ = "RVP"

    params: Any = None
    soil_classes: Optional[rc.SoilClasses] = optfield(alias="SoilClasses")
    soil_profiles: Optional[rc.SoilProfiles] = optfield(alias="SoilProfiles")
    vegetation_classes: Optional[rc.VegetationClasses] = optfield(alias="VegetationClasses")
    land_use_classes: Optional[rc.LandUseClasses] = optfield(alias="LandUseClasses")
    terrain_classes: Optional[rc.TerrainClasses] = optfield(alias="TerrainClasses")
    soil_parameter_list: Optional[rc.SoilParameterList] = optfield(alias="SoilParameterList")
    land_use_parameter_list: Optional[rc.LandUseParameterList] = optfield(alias="LandUseParameterList")
    vegetation_parameter_list: Optional[rc.VegetationParameterList] = optfield(alias="VegetationParameterList")

    channel_profile: Optional[Sequence[rc.ChannelProfile]] = optfield(alias="ChannelProfile")

    # TODO: create list of all available parameters to constrain key
    global_parameter: Optional[dict[str, Sym]] = Field({}, alias="GlobalParameter")
    rain_snow_transition: Optional[rc.RainSnowTransition] = optfield(alias="RainSnowTransition")
    seasonal_relative_lai: Optional[rc.SeasonalRelativeLAI] = optfield(alias="SeasonalRelativeLAI")
    seasonal_relative_height: Optional[rc.SeasonalRelativeHeight] = optfield(alias="SeasonalRelativeHeight")


class RVC(RV):
    __rv__ = "RVC"

    hru_state_variable_table: Optional[rc.HRUStateVariableTable] = optfield(alias="HRUStateVariableTable")
    basin_state_variables: Optional[rc.BasinStateVariables] = optfield(alias="BasinStateVariables")
    uniform_initial_conditions: Optional[dict[str, Sym]] = optfield(alias="UniformInitialConditions")


class RVH(RV):
    __rv__ = "RVH"

    sub_basins: Optional[rc.SubBasins] = optfield(alias="SubBasins")
    sub_basin_group: Optional[Sequence[rc.SubBasinGroup]] = optfield(alias="SubBasinGroup")
    sub_basin_properties: Optional[rc.SubBasinProperties] = optfield(alias="SubBasinProperties")
    sb_group_property_multiplier: Optional[Sequence[rc.SBGroupPropertyMultiplier]] = optfield(alias="SBGroupPropertyMultiplier")
    hrus: Optional[rc.HRUs] = optfield(alias="HRUs")
    hru_group: Optional[Sequence[rc.HRUGroup]] = optfield(alias="HRUGroup")
    reservoirs: Optional[Sequence[rc.Reservoir]] = optfield(alias="Reservoirs")


class RVE(RV):
    __rv__ = "RVE"

    enkf_mode: Optional[o.EnKFMode] = optfield(alias="EnKFMode")
    window_size: Optional[int] = optfield(alias="WindowSize")
    solution_run_name: Optional[str] = optfield(alias="SolutionRunName")
    extra_rvt_filename: Optional[str] = optfield(alias="ExtraRVTFilename")
    output_directory_format: Optional[Union[str, Path]] = optfield(alias="OutputDirectoryFormat")
    forecast_rvt_filename: Optional[str] = optfield(alias="ForecastRVTFilename")
    truncate_hindcasts: Optional[bool] = optfield(alias="TruncateHindcasts")
    forcing_perturbation: Optional[Sequence[rc.ForcingPerturbation]] = optfield(alias="ForcingPerturbation")
    assimilated_state: Optional[Sequence[rc.AssimilatedState]] = optfield(alias="AssimilatedState")
    assimilate_streamflow: Optional[Sequence[rc.AssimilateStreamflow]] = optfield(alias="AssimilateStreamflow")
    observation_error_model: Optional[Sequence[rc.ObservationErrorModel]] = optfield(alias="ObservationErrorModel")


class Config(RVI, RVC, RVH, RVT, RVP, RVE):
    __rv__ = None

    @staticmethod
    def header(rv):
        """Return the header to print at the top of each RV file."""
        import ravenpy

        return dedent(
            f"""
        ###########################################################################################################
        :FileType          {rv.upper()} Raven {__raven_version__}
        :WrittenBy         RavenPy {ravenpy.__version__} based on setups provided by James Craig and Juliane Mai
        :CreationDate      {dt.datetime.now().isoformat(timespec="seconds")}
        #----------------------------------------------------------------------------------------------------------
        \n\n
        """
        )

    @field_validator("params", mode="before")
    @classmethod
    def _cast_to_dataclass(cls, data: Union[dict, Sequence]):
        """Cast params to a dataclass."""
        # Needed because pydantic v2 does not cast tuples automatically.
        if data is not None:
            if is_dataclass(data):
                return data

            if isinstance(data, dict):
                return cls.model_fields["params"].annotation(**data)

            return cls.model_fields["params"].annotation(*data)

    @field_validator("global_parameter", mode="before")
    @classmethod
    def _update_defaults(cls, v, info: ValidationInfo):
        """Some configuration parameters should be updated with user given arguments, not overwritten."""
        return {**cls.model_fields[info.field_name].default, **v}

    def set_params(self, params: Union[dict, Sequence]) -> "Config":
        """Return a new instance of Config with params frozen to their numerical values."""
        # Create params with numerical values
        if not self.is_symbolic:
            raise ValueError(
                "Setting `params` on a configuration without symbolic expressions has no effect. "
                "Leave `params` to its default value when instantiating the emulator configuration."
            )

        p = self.__class__._cast_to_dataclass(params)

        # Parse symbolic expressions using numerical params values
        out = parse_symbolic(self.__dict__, **asdict(p))
        out["params"] = p

        # Instantiate config class
        return self.__class__.model_construct(**out)

    def set_solution(self, fn: Path, timestamp: bool = True) -> "Config":
        """
        Return a new instance of Config with hru, basin states and start date set from an existing solution.

        Parameters
        ----------
        fn : Path
           Path to solution file.
        timestamp : bool
           If False, ignore time stamp information in the solution. If True, the solution
           will set StartDate to the solution's timestamp.

        Returns
        -------
        Config
            Config with internal state set from the solution file.
        """
        from .defaults import CALENDAR
        from .parsers import parse_solution

        try:
            calendar = self.calendar.value
        except AttributeError:
            calendar = CALENDAR

        out = self.__dict__.copy()
        sol = parse_solution(fn, calendar=calendar)
        if timestamp is False:
            sol.pop("start_date")

        out.update(**sol, uniform_initial_conditions={})

        return self.__class__(**out)

    def duplicate(self, **kwds):
        """Duplicate this model, changing the values given in the keywords."""
        out = self.model_copy(deep=True)
        for key, val in self.model_validate(kwds).model_dump(exclude_unset=True).items():
            setattr(out, key, val)
        return out

    def _rv(self, rv: str):
        """Return RV configuration."""
        import inspect

        # if self.is_symbolic:
        #     raise ValueError(
        #         "Cannot write RV files if `params` has symbolic variables. Use `set_params` method to set numerical "
        #         "values for `params`."
        #     )
        # Get RV class
        for cls in inspect.getmro(self.__class__):
            if cls.__rv__ == rv.upper():
                break

        # Instantiate RV class
        attrs = dict(self)
        rv_attrs = {f: attrs[f] for f in cls.model_fields}

        # Get model parameters and convert symbolic expressions.
        if self.params is not None:
            p = asdict(self.params)
            rv_attrs = parse_symbolic(rv_attrs, **p)

        rv = cls.model_validate(rv_attrs)
        return rv.to_rv()

    @property
    def is_symbolic(self):
        """Return True if configuration contains symbolic expressions."""
        if self.params is not None:
            p = {field.name: getattr(self.params, field.name) for field in fields(self.params)}
            return is_symbolic(p)

        return False

    @property
    def rvi(self):
        return self._rv("rvi")

    @property
    def rvt(self):
        return self._rv("rvt")

    @property
    def rvp(self):
        return self._rv("rvp")

    @property
    def rvc(self):
        return self._rv("rvc")

    @property
    def rvh(self):
        return self._rv("rvh")

    @property
    def rve(self):
        return self._rv("rve")

    def write_rv(
        self,
        workdir: Union[str, Path],
        modelname: Optional[str] = None,
        overwrite: bool = False,
        header: bool = True,
    ):
        """
        Write configuration files to disk.

        Parameters
        ----------
        workdir : str, Path
            A directory where rv files will be written to disk.
        modelname : str
            File name stem for rv files. If not given, defaults to `RunName` if set, otherwise `raven`.
        overwrite : bool
            If True, overwrite existing configuration files.
        header : bool
            If True, write a header at the top of each RV file.
        """
        workdir = Path(workdir)
        if not workdir.exists():
            workdir.mkdir(parents=True)

        mandatory = ["rvi", "rvp", "rvc", "rvh", "rvt"]
        optional = ["rve"]

        if modelname is None:
            modelname = self.run_name or "raven"

        out = {}
        for rv in mandatory + optional:
            fn = workdir / f"{modelname}.{rv}"
            if fn.exists() and not overwrite:
                raise OSError(f"{fn} already exists and would be overwritten.")

            # RV header
            text = self.header(rv) if header else ""

            # RV content
            text += self._rv(rv)

            # Write to disk
            if rv in mandatory or text.strip():
                fn.write_text(text)
                out[rv] = fn

        return out

    def zip(
        self,
        workdir: Union[str, Path],
        modelname: Optional[str] = None,
        overwrite: bool = False,
    ):
        """
        Write configuration to zip file.

        Parameters
        ----------
        workdir : Path, str
            Path to zip archive storing RV files.
        modelname : str, optional
            File name stem for rv files. If not given, defaults to `RunName` if set, otherwise `raven`.
        overwrite : bool
            If True, overwrite existing configuration zip file.
        """
        workdir = Path(workdir)
        if not workdir.exists():
            workdir.mkdir(parents=True)

        if modelname is None:
            modelname = self.run_name or "raven"

        fn = workdir / f"{modelname}.zip"
        if fn.exists() and not overwrite:
            raise OSError(f"{fn} already exists and would be overwritten.")

        with zipfile.ZipFile(fn, "w") as fh:
            for rv in ["rvi", "rvp", "rvc", "rvh", "rvt", "rve"]:
                f = f"{modelname}.{rv}"
                txt = self._rv(rv)
                if txt.strip():
                    fh.writestr(f, txt)
        return zip


def is_symbolic(params: Union[dict, Any]) -> bool:
    """Return True if parameters include a symbolic variable."""
    from pymbolic.primitives import Variable

    if is_dataclass(params) and not isinstance(params, type):
        params = {field.name: getattr(params, field.name) for field in fields(params)}

    return any([isinstance(v, Variable) for v in params.values()])



import re
from collections.abc import Sequence
from enum import Enum
from textwrap import dedent, indent
from typing import Any, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, RootModel, model_validator
from pymbolic.primitives import ExpressionNode, Variable


"""
Notes
-----
In pydantic, BaseModel and dataclasses behave slightly differently.
A dataclass object can be instantiated using positional arguments, while BaseModels cannot.
This is why Params is defined as a dataclass::

    Two cases
    RV
        Command, alias="A1"
        Sequence[Command], alias="A2"

Sometimes, we want A2 to be printed as a command (HydrologicProcesses)
Sometime, we don't (Gauges)
In general, let's print the alias except if the parent is __root__, or if the parent's name is the same as the
children (Data holding multiple Data elements).

"""


SymConfig = ConfigDict(arbitrary_types_allowed=True)


Sym = Union[Variable, ExpressionNode, float, None]


def optfield(**kwds):
    """Shortcut to create an optional field with an alias."""
    return Field(**kwds, default_factory=lambda: None, validate_default=False)


class Params:
    pass


def encoder(v: dict) -> dict:
    r"""
    Return string representation of objects in dictionary.

    This is meant to be applied to BaseModel attributes that either have an `alias` defined,
    or have a `root` attribute. The objective is to avoid creating `Command` objects for every
    configuration option:

        - bool: ':{cmd}\n' if obj else ''
        - dict: ':{cmd} {key} {value}'
        - enum: ':{cmd} {obj.value}'
        - Command: obj.to_rv()
        - Sequence: complicated
        - Any other: ':{cmd} {obj}'
    """
    for cmd, obj in v.items():
        if obj is None:
            continue
        elif isinstance(obj, bool):
            # :Command\n
            out = f":{cmd}\n" if obj else ""
        elif isinstance(obj, dict):
            # :Command key value\n
            out = "\n".join([f":{cmd} {key} {value}" for (key, value) in obj.items()]) + "\n\n"
        elif isinstance(obj, Enum):
            # :Command value\n
            out = f":{cmd:<20} {obj.value}\n"

        elif hasattr(obj, "to_rv"):
            # Custom
            out = obj.to_rv()
        elif issubclass(obj.__class__, _Record):
            out = str(obj)
        elif isinstance(obj, (list, tuple)):
            s = map(str, obj)
            o0 = obj[0]
            if cmd == "root" or issubclass(o0.__class__, FlatCommand):
                out = "".join(s)
            # elif cmd == "Attributes":
            #     out = f":{cmd}," + ",".join(s) + "\n"
            elif cmd in ["Parameters", "Attributes", "Units"]:
                fmt = ":{cmd:<15}" + len(obj) * ",{:>18}" + "\n"
                out = fmt.format(*obj, cmd=cmd)
            elif issubclass(o0.__class__, (_Command, _Record)):
                rec = indent("\n".join(s), o0._indent)
                out = f":{cmd}\n{rec}\n:End{cmd}\n"
            elif isinstance(o0, Enum):
                seq = " ".join([o.value for o in obj])
                out = f":{cmd:<20} {seq}\n"
            else:
                seq = " ".join(s)
                out = f":{cmd:<20} {seq}\n"
        else:
            out = f":{cmd:<20} {obj}\n"

        v[cmd] = str(out)

    return v


class _Record(BaseModel):
    pass


class Record(_Record):
    """
    A Record has no nested Command or Record objects. It is typically a list of named
    values on a single line.

    For example, SubBasins is a ListCommand, whose root is a list of `SubBasin` Records.
    """

    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True, populate_by_name=True)


class RootRecord(RootModel, _Record):
    """
    A Record with a root attribute. This is typically used for an unnamed list of
    values on a single line.

    For example, the list of HRUs in an HRUGroup is a RootRecord, and the weights of a GridWeights
    command are a sequence of records.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]

    def __str__(self):
        return " ".join(map(str, self.root))


class _Command(BaseModel):
    """Base class for Raven commands."""

    @property
    def _indent(self):
        return "  "

    @property
    def _template(self):
        return """
               :{_cmd}
               {_commands}{_records}
               :End{_cmd}
               """

    def __str__(self):
        return self.to_rv()

    def __subcommands__(self) -> tuple[dict[str, str], list]:
        """Return dictionary of class attributes that are Raven models."""
        cmds = {}
        recs = []
        cls = self.__class__
        for key, field in cls.model_fields.items():
            obj = self.__dict__[key]
            if obj is not None:
                if issubclass(obj.__class__, _Record):
                    recs = [obj]
                elif issubclass(obj.__class__, _Command):
                    cmds[field.alias] = self.__dict__[key]
                else:
                    try:
                        o = obj[0]
                    except (TypeError, IndexError, KeyError):
                        o = None
                    if issubclass(o.__class__, _Record):
                        recs = obj
                    elif field.alias is not None:
                        cmds[field.alias] = self.__dict__[key]
                    elif key == "root":
                        cmds[key] = self.__dict__[key]
                # elif (
                #     getattr(field.annotation, "_name", "") == "Sequence"
                #     and len(obj)
                #     and issubclass(obj[0].__class__, Record)
                # ):
                #     recs = obj
                # elif field.has_alias or field.alias == "__root__":
                #     cmds[field.alias] = self.__dict__[key]

        for key in self.__private_attributes__.keys():
            cmds[key.strip("_")] = getattr(self, key)

        return cmds, recs

    def to_rv(self):
        """Return Raven configuration string."""
        cmds, recs = self.__subcommands__()

        d = self.model_dump()
        if not isinstance(d, dict):
            d = dict(root=d)

        # Write command strings
        d.update(encoder(cmds))

        # Write record strings
        recs = "\n".join(map(str, recs))
        if recs:
            recs = indent(recs, self._indent)

        d["_cmd"] = self.__class__.__name__
        d["_commands"] = indent("".join(cmds.values()).strip(), self._indent)
        if cmds and recs and "_commands" in self._template:
            recs = "\n" + recs
        d["_records"] = recs
        return dedent(self._template).format(**d)


class Command(_Command):
    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True, populate_by_name=True)


class RootCommand(RootModel, _Command):
    """Generic Command for root models."""

    root: Any
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)


class FlatCommand(Command):
    """
    Only used to discriminate Commands that should not be nested.

    HRUGroup, ReadFromNetCDF, Reservoir are examples of FlatCommand.
    """


class LineCommand(FlatCommand):
    r"""
    A non-nested Command on a single line.

    :CommandName {field_1} {field_2} ... {field_n}\n

    EvaluationPeriod is a FlatCommand.
    """

    def to_rv(self):
        cls = self.__class__
        out = [f":{cls.__name__:<20}"]
        for field in cls.model_fields.keys():
            out.append(str(getattr(self, field)))  # noqa: PERF401

        return " ".join(out) + "\n"

    @classmethod
    def parse(cls, s):
        """Parse the command and return an instance of LineCommand."""
        pat = rf":{cls.__name__}\s+(?P<args>.+)"
        fields = cls.model_fields.keys()
        if match := re.match(pat, s):
            args = match.group("args").split()
            return cls(**dict(zip(fields, args)))


class ListCommand(RootModel, _Command):
    """Use so that commands with __root__: Sequence[Record] behave like a list."""

    root: Sequence[Any]

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]

    def __len__(self):
        return len(self.root)

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)


class ParameterList(Record):
    name: str = ""
    values: Sequence[Union[Sym, str, None]] = ()

    @model_validator(mode="before")
    @classmethod
    def no_none_in_default(cls, data):
        """Make sure that no values are None for the [DEFAULT] record."""
        if data["name"] == "[DEFAULT]" and None in data["values"]:
            raise ValueError("Default record can not contain None.")
        return data

    def __str__(self):
        fmt = "{name:<16}" + len(self.values) * ",{:>18}"
        evals = []
        for v in self.values:
            ev = "_DEFAULT" if v is None else v
            evals.append(ev)

        return fmt.format(*evals, name=self.name)


class GenericParameterList(Command):
    parameters: Sequence[str] = Field(alias="Parameters", description="Parameter names")
    units: Optional[Sequence[str]] = Field(None, alias="Units")
    pl: Sequence[ParameterList]

    @model_validator(mode="after")
    def num_values_equal_num_names(self):
        """Check that the length of the parameter list equals the number of given parameter names."""
        n = len(self.parameters)

        for pl in self.pl:
            if len(pl.values) != n:
                raise ValueError("Number of values should match number of parameters.")

        return self

    @model_validator(mode="after")
    def set_default_units(self):
        n = len(self.parameters)
        if self.units is None:
            self.units = ("none",) * n
        return self


class RV(Command):
    """Base class for RV configuration objects."""

    __rv__ = ""

    @property
    def _template(self):
        return "{_commands}\n"

    @property
    def _indent(self):
        return ""

    model_config = ConfigDict(populate_by_name=True, validate_default=True, validate_assignment=True)


def parse_symbolic(value, **kwds):
    """
    Inject values of symbolic variables into object and return object.

    Note that parsing the output of `model_dump` can cause problems because there is not always enough information in the
    dictionary to recreate the correct model.
    """
    from pymbolic.mapper.evaluator import EvaluationMapper
    from pymbolic.primitives import ExpressionNode, Variable

    if isinstance(value, dict):
        return {k: parse_symbolic(v, **kwds) for k, v in value.items()}

    elif isinstance(value, (list, tuple)):
        return type(value)(parse_symbolic(v, **kwds) for v in value)

    elif isinstance(value, (_Record, _Command)):
        attrs = value.__dict__
        if isinstance(value, RootModel):
            attrs = attrs["root"]
        return value.model_validate(parse_symbolic(attrs, **kwds))

    elif isinstance(value, (Variable, ExpressionNode)):
        # Inject numerical values numerical value
        return EvaluationMapper(context=kwds)(value)

    else:
        return value



from collections.abc import Sequence
from dataclasses import field, make_dataclass
from typing import Union

from pydantic import Field, field_validator
from pydantic.dataclasses import dataclass
from pymbolic.primitives import Variable

import ravenpy.config.processes as p
from ravenpy.config import commands as rc
from ravenpy.config import options as o
from ravenpy.config.base import Sym, SymConfig
from ravenpy.config.commands import (
    HRU,
    PL,
    LandUseClasses,
    LandUseParameterList,
    SoilClasses,
    SoilParameterList,
    SoilProfiles,
    VegetationClasses,
)
from ravenpy.config.defaults import nc_attrs
from ravenpy.config.rvs import Config


P = dataclass(
    make_dataclass(
        "Params",
        [(f"X{i:02}", Sym, field(default=Variable(f"X{i:02}"))) for i in range(1, 22)],
    ),
    config=SymConfig,
)

# Bug: RavenC tries to write two variables with the same name `Snow Melt (Liquid) [mm]` to netCDF.


class LandHRU(HRU):
    land_use_class: str = "FOREST"
    veg_class: str = "FOREST"
    soil_profile: str = "DEFAULT_P"
    aquifer_profile: str = "[NONE]"
    terrain_class: str = "[NONE]"


class HRUs(rc.HRUs):
    """
    HRUs command for GR4J.

    Pydantic is able to automatically detect if an HRU is Land or Lake if `hru_type` is provided.
    """

    root: Sequence[LandHRU]


class SACSMA(Config):
    """
    Sacramento - Soil Moisture Accounting model

    References
    ----------
    Sorooshian, S., Duan, Q., and Gupta, V. K. (1993), Calibration of rainfall-runoff models:
    Application of global optimization to the Sacramento Soil Moisture Accounting Model,
    Water Resour. Res., 29( 4), 1185– 1194, doi:10.1029/92WR02617.
    """

    params: P = P()
    hrus: HRUs = Field([LandHRU()], alias="HRUs")
    netcdf_attribute: dict[str, str] = {"model_id": "SACSMA"}
    sub_basins: rc.SubBasins = Field([rc.SubBasin()], alias="SubBasins")
    write_netcdf_format: bool = Field(True, alias="WriteNetcdfFormat")
    time_step: Union[float, str] = Field(1.0, alias="TimeStep")
    calendar: o.Calendar = Field("PROLEPTIC_GREGORIAN", alias="Calendar")
    routing: o.Routing = Field("ROUTE_NONE", alias="Routing")
    catchment_route: o.CatchmentRoute = Field("ROUTE_GAMMA_CONVOLUTION", alias="CatchmentRouting")
    evaporation: o.Evaporation = Field(o.Evaporation.OUDIN, alias="Evaporation")
    rain_snow_fraction: o.RainSnowFraction = Field(o.RainSnowFraction.DATA, alias="RainSnowFraction")
    potential_melt_method: o.PotentialMeltMethod = Field(o.PotentialMeltMethod.DEGREE_DAY, alias="PotentialMeltMethod")
    soil_model: rc.SoilModel = Field(7, alias="SoilModel")

    hydrologic_processes: Sequence[Union[rc.Process, p.Conditional]] = Field(
        [
            p.SnowBalance(algo="SNOBAL_SIMPLE_MELT", source="SNOW", to="PONDED_WATER"),
            p.Precipitation(algo="RAVEN_DEFAULT", source="ATMOS_PRECIP", to="MULTIPLE"),
            p.SoilEvaporation(algo="SOILEVAP_SACSMA", source="MULTIPLE", to="ATMOSPHERE"),
            p.SoilBalance(algo="SOILBAL_SACSMA", source="MULTIPLE", to="MULTIPLE"),
            p.OpenWaterEvaporation(algo="OPEN_WATER_RIPARIAN", source="SURFACE_WATER", to="ATMOSPHERE"),
        ],
        alias="HydrologicProcesses",
    )

    soil_classes: SoilClasses = Field(
        [
            {"name": "UZT"},
            {"name": "UZF"},
            {"name": "LZT"},
            {"name": "LZFP"},
            {"name": "LZFS"},
            {"name": "ADIM"},
            {"name": "GW"},
        ],
        alias="SoilClasses",
    )
    vegetation_classes: VegetationClasses = Field(
        [{"name": "FOREST", "max_ht": 4, "max_lai": 5, "max_leaf_cond": 5}],
        alias="VegetationClasses",
    )
    land_use_classes: LandUseClasses = Field(
        [rc.LU(name="FOREST", impermeable_frac=P.X13, forest_coverage=1.0)],
        alias="LandUseClasses",
    )

    soil_profiles: SoilProfiles = Field(
        [
            {"name": "LAKE"},
            {
                "name": "DEFAULT_P",
                "soil_classes": ("UZT", "UZF", "LZT", "LZFP", "LZFS", "ADIM", "GW"),
                "thicknesses": (P.X04, P.X05, P.X06, P.X08, P.X07, 100, 100),
            },
        ],
        alias="SoilProfiles",
    )

    soil_parameter_list: SoilParameterList = Field(
        {
            "parameters": (
                "POROSITY",
                "SAC_PERC_ALPHA",
                "SAC_PERC_EXPON",
                "SAC_PERC_PFREE",
                "BASEFLOW_COEFF",
                "UNAVAIL_FRAC",
            ),
            "pl": [
                PL(
                    name="[DEFAULT]",
                    values=(1.0, P.X11, P.X10, P.X09, 0, 0),
                ),
                PL(
                    name="UZF",
                    values=(1.0, P.X11, P.X10, P.X09, P.X03, 0),
                ),
                PL(
                    name="LZFP",
                    values=(1.0, P.X11, P.X10, P.X09, P.X01, P.X16),
                ),
                PL(
                    name="LZFS",
                    values=(1.0, P.X11, P.X10, P.X09, P.X02, 0),
                ),
            ],
        },
        alias="SoilParameterList",
    )
    land_use_parameter_list: LandUseParameterList = Field(
        {
            "parameters": (
                "GAMMA_SHAPE",
                "GAMMA_SCALE",
                "MELT_FACTOR",
                "STREAM_FRACTION",
                "MAX_SAT_AREA_FRAC",
                "BF_LOSS_FRACTION",
            ),
            "pl": [
                PL(
                    name="[DEFAULT]",
                    values=(
                        P.X18,
                        P.X19,
                        P.X17,
                        P.X15,
                        P.X14,
                        P.X12,
                    ),
                )
            ],
        },
        alias="LandUseParameterList",
    )
    vegetation_parameter_list: rc.VegetationParameterList = Field(
        {
            "parameters": ("RAIN_ICEPT_PCT", "SNOW_ICEPT_PCT"),
            "pl": [PL(name="[DEFAULT]", values=(0, 0))],
        },
        alias="VegetationParameterList",
    )
    hru_state_variable_table: rc.HRUStateVariableTable = Field(
        [
            rc.HRUState(
                hru_id=1,
                data={
                    "SOIL[0]": P.X04 * 1000,
                    "SOIL[2]": P.X06 * 1000,
                },
            ),
        ],
        alias="HRUStateVariableTable",
    )
    _nc_attrs = field_validator("netcdf_attribute")(nc_attrs)

    def __init__(self, **data):
        super().__init__(**data)

        if self.gauge:
            for gauge in self.gauge:
                gauge.rain_correction = self.params.X20
                gauge.snow_correction = self.params.X21



# Taken from https://cfconventions.org/Data/cf-standard-names/current/build/cf-standard-name-table.html
# Augmented with https://pcmdi.llnl.gov/mips/cmip3/variableList.html
CF_RAVEN = {
    "TEMP_MIN": ("tasmin",),
    "TEMP_DAILY_MIN": ("tasmin",),
    "TEMP_MAX": ("tasmax",),
    "TEMP_DAILY_MAX": ("tasmax",),
    "TEMP_AVE": ("tas", "air_temperature"),
    "TEMP_DAILY_AVE": ("tas", "air_temperature"),
    "RAINFALL": (
        "rainfall_flux",
        "prra",
        "rainfall",
    ),
    "PRECIP": ("pr", "precipitation_flux"),
    "PRECIP_DAILY_AVE": ("pr", "precipitation_flux"),
    "SNOWFALL": ("prsn", "snowfall_flux", "snowfall"),
    "REL_HUMIDITY": ("hurs", "relative_humidity"),
    "AIR_PRES": ("ps", "surface_air_pressure"),
    "SHORTWAVE": ("rsds", "surface_downwelling_shortwave_flux_in_air"),
    "LW_INCOMING": ("rlds", "surface_downwelling_longwave_flux_in_air"),
    "CLOUD_COVER": ("clt", "cloud_area_fraction"),
    "WIND_VEL": (
        "sfcWind",
        "wind_speed",
    ),
    "PET": ("evspsbl", "water_evapotranspiration_flux", "water_evaporation_flux"),
    "HYDROGRAPH": ("water_volume_transport_in_river_channel",),
}


NetCDFAttribute = {
    "title": "Simulated river discharge",
    "history": "Created on {dt.datetime.now().strftime('%Y-%m-%d %H:%M:%S')} by Raven",
    "references": "Craig, J.R., and the Raven Development Team, Raven user's and developer's manual ("
    "Version 2.8), URL: http://raven.uwaterloo.ca/ (2018).",
    "comment": "Raven Hydrological Framework version {raven_version}",
    "model_id": "{identifier}",
    "time_frequency": "day",
    "time_coverage_start": "{start_date}",
    "time_coverage_end": "{end_date}",
}

MonthlyAverages = {
    "TEMP_AVE": "MonthlyAveTemperature",
    "TEMP_MIN": "MonthlyMinTemperature",
    "TEMP_MAX": "MonthlyMaxTemperature",
    "PET": "MonthlyAveEvaporation",
}

RAVEN_OUTPUT_FMT = {
    "solution": "{run_name}solution.rvc",
    "hydrograph": "{run_name}Hydrographs.nc",
    "storage": "{run_name}WatershedStorage.nc",
    "diagnostics": "{run_name}Diagnostics.csv",
    "messages": "Raven_errors.txt",
}



"""
Library to perform graphs for the streamflow time series analysis.

The following graphs can be plotted:
    - hydrograph
    - mean_annual_hydrograph
    - spaghetti_annual_hydrograph
"""

from collections.abc import Sequence
from pathlib import Path
from typing import Union

import matplotlib.pyplot
import numpy as np
import pandas as pd
import xarray as xr
from matplotlib import pyplot as plt
from scipy import stats
from xclim.core.units import units2pint

from ravenpy.utilities.mk_test import mk_test_calc


# TODO: Review docstrings, ensure numpydoc convention compliance


def hydrograph(file_list: Sequence[Union[str, Path]]):
    """
    Create a graphic of the hydrograph for each model simulation.

    Parameters
    ----------
    file_list : Sequence of str or Path
        Raven output files containing simulated streamflows.
    """
    ds = [xr.open_dataset(file) for file in file_list]

    # Get time data for the plot
    dates = pd.DatetimeIndex(ds[0].time.values)
    first_date = dates.min().strftime("%Y/%m/%d")
    last_date = dates.max().strftime("%Y/%m/%d")

    basin_name = ds[0].basin_name.values[0]  # selected basin name

    fig, ax = plt.subplots()  # initialize figure

    # Plot the observed streamflows if available
    if hasattr(ds[0], "q_obs"):
        q_obs = ds[0].q_obs
        plt.plot(dates, q_obs, linewidth=2, label="obs")

    # Plot the simulated streamflows for each hydrological model
    q_sim = [h.q_sim for h in ds]
    for sim in q_sim:
        plt.plot(dates, sim, linewidth=2, label="sim: " + basin_name)

    # plt.xlim([first_date, last_date])
    plt.ylim(bottom=0, top=None)
    ax.set_xlabel("Time")
    ax.set_ylabel(r"$Streamflow [m^3s^{{-1}}]$")
    ax.set_title(f"Hydrograph between {first_date} and {last_date}\nSelected basin: {basin_name} basin.")
    ax.legend()
    ax.grid()

    plt.xticks(rotation=90)
    plt.tight_layout()

    return fig


def mean_annual_hydrograph(file_list: Sequence[Union[str, Path]]):
    """
    Create a graphic of the mean hydrological cycle for each model simulation.

    Parameters
    ----------
    file_list : Sequence of str or Path
        Raven output files containing simulated streamflows.
    """
    # Time series for the plot
    ds = [xr.open_dataset(file) for file in file_list]

    # Get time data for the plot
    dates = pd.DatetimeIndex(ds[0].time.values)
    first_date = dates.min().strftime("%Y/%m/%d")
    last_date = dates.max().strftime("%Y/%m/%d")

    basin_name = ds[0].basin_name.values[0]  # selected basin name

    fig, ax = plt.subplots()  # initialize figure

    # Plot the observed streamflows if available
    mah_obs = None
    if hasattr(ds[0], "q_obs"):
        q_obs = ds[0].q_obs
        mah_obs = q_obs.groupby("time.dayofyear").mean()

        plt.plot(mah_obs.dayofyear, mah_obs, linewidth=2, label="obs")

    # Plot the simulated streamflows for each hydrological model
    q_sim = [h.q_sim for h in ds]
    mahs_sim = [q.groupby("time.dayofyear").mean() for q in q_sim]

    for mah in mahs_sim:
        plt.plot(mah.dayofyear, mah, linewidth=2, label="sim: " + basin_name)

    plt.xticks(
        np.linspace(0, 365, 13)[:-1],
        (
            "Jan",
            "Feb",
            "Mar",
            "Apr",
            "May",
            "Jun",
            "Jul",
            "Aug",
            "Sep",
            "Oct",
            "Nov",
            "Dec",
        ),
    )

    if mah_obs is not None:
        plt.xlim(0, mah_obs.shape[0])
    plt.ylim(bottom=0, top=None)
    ax.set_xlabel("Time")
    ax.set_ylabel(r"$Streamflow [m^3s^{{-1}}]$")
    ax.set_title(f"Hydrograph between {first_date} and {last_date}\nSelected basin: {basin_name} basin.")
    ax.legend()
    ax.grid()

    plt.tight_layout()

    return fig


def spaghetti_annual_hydrograph(file: Union[str, Path]):
    """
    Create a spaghetti plot of the mean hydrological cycle for one model simulations.

    The mean simulation is also displayed.

    Parameters
    ----------
    file : str or Path
        Raven output files containing simulated streamflows of one model.
    """
    # Time series for the plot
    ds = xr.open_dataset(file)

    # Get time data for the plot
    dates = pd.DatetimeIndex(ds.time.values)
    first_date = dates.min().strftime("%Y/%m/%d")
    last_date = dates.max().strftime("%Y/%m/%d")

    basin_name = ds.basin_name.values[0]  # selected basin name

    fig, ax = plt.subplots()  # initialize figure

    # Plot the observed streamflows if available
    if hasattr(ds, "q_obs"):
        q_obs = ds.q_obs
        mah_obs = q_obs.groupby("time.year")
        mah_obs_mean = q_obs.groupby("time.dayofyear").mean()

        for year in mah_obs.groups.keys():
            plt.plot(
                np.arange(1, q_obs.values[mah_obs.groups[year]].shape[0] + 1, 1),
                q_obs.values[mah_obs.groups[year]],
                linewidth=1,
                color="C0",
            )

        plt.plot(mah_obs_mean.dayofyear, mah_obs_mean, linewidth=2, color="C0", label="obs")

    # Plot the simulated streamflows for each hydrological model
    q_sim = ds.q_sim
    mah_sim = q_sim.groupby("time.year")
    mah_sim_mean = q_sim.groupby("time.dayofyear").mean()

    for year in mah_sim.groups.keys():
        plt.plot(
            np.arange(1, q_sim.values[mah_sim.groups[year]].shape[0] + 1, 1),
            q_sim.values[mah_sim.groups[year]],
            linewidth=1,
            color="C1",
        )

    plt.plot(
        mah_sim_mean.dayofyear,
        mah_sim_mean,
        linewidth=2,
        color="C1",
        label="sim: " + "<model_name>",
    )

    plt.xticks(
        np.linspace(0, 365, 13)[:-1],
        (
            "Jan",
            "Feb",
            "Mar",
            "Apr",
            "May",
            "Jun",
            "Jul",
            "Aug",
            "Sep",
            "Oct",
            "Nov",
            "Dec",
        ),
    )

    plt.xlim(0, mah_sim_mean.shape[0])
    plt.ylim(bottom=0, top=None)
    ax.set_xlabel("Time")
    ax.set_ylabel(r"$Streamflow [m^3s^{{-1}}]$")
    ax.set_title(f"Spaghetti annual hydrograph between {first_date} and {last_date}\n Selected basin: {basin_name} basin.")
    ax.legend()
    ax.grid()

    plt.tight_layout()

    return fig


def ts_graphs(file, trend: bool = True, alpha: float = 0.05):
    """
    Create a figure with the statistics so one can see a trend in the data.

    Graphs for time series statistics.

    Parameters
    ----------
    file : str or Path
        xarray-compatible file containing streamflow statistics for one run.
    trend : bool
        If True, the slope will be plotted.
    alpha : float
        Significance level for the Mann-Kendall test.
    """
    # Time series for the plot
    ds = xr.open_dataset(file)
    titlename = ds["ts_stats"].description
    values = ds["ts_stats"].values[:]

    values[0] = 2
    values[1] = 1
    values[2] = 3
    values[3] = 3
    values[4] = 3
    values[5] = 5
    values[6] = 6
    values[7] = 7
    values[8] = 10
    values[9] = 4

    # Get time data for the plot
    dates = pd.DatetimeIndex(ds.time.values)
    # first_date = dates.min()#.strftime('%Y/%m/%d')
    # last_date = dates.max()#.strftime('%Y/%m/%d')
    res = None

    if trend:
        res = stats.theilslopes(values, dates)
        trd, h, p, z = mk_test_calc(values, alpha=alpha)
        titlename = titlename + ", Mann-Kendall h=" + str(h) + ", p-value=" + str(np.round(p, 4))

    fig, ax = plt.subplots()
    ax.plot(dates, values, label="time-series index")

    # plt.xlim([first_date, last_date])
    ax.set_xlabel("Time")
    ax.set_ylabel(r"$Streamflow [m^3s^{{-1}}]$")

    ax.set_title(titlename)

    ax.grid()
    plt.tight_layout()

    if trend:
        # TODO: This does not work yet, trying to compute the y-value of a datetime x-axis value * a slope...
        ax.plot(
            dates,
            res[1] + res[0] * dates,
            linestyle="--",
            linewidth=2,
            label="Sen's Slope",
        )

    ax.legend()

    return fig


def ts_fit_graph(ts: xr.DataArray, params: xr.DataArray) -> matplotlib.pyplot.Figure:
    """
    Create graphic showing a histogram of the data and the distribution fitted to it.

    The graphic contains one panel per watershed.

    Parameters
    ----------
    ts : xr.DataArray
        Stream flow time series with dimensions (time, nbasins).
    params : xr.DataArray
        Fitted distribution parameters returned by `xclim.land.fit` indicator.

    Returns
    -------
    matplotlib.pyplot.Figure
        Figure showing a histogram and the parameterized pdf.
    """
    from xclim.indices.stats import get_dist

    n = ts.nbasins.size
    dist = params.attrs["scipy_dist"]

    fig, axes = plt.subplots(n, figsize=(10, 6), squeeze=False)
    if params.isnull().any():
        raise ValueError("Null values found in `params`.")

    for i in range(n):
        ax = axes.flat[i]
        ax2 = plt.twinx(ax)
        p = params.isel(nbasins=i)
        t = ts.isel(nbasins=i).dropna(dim="time")

        # Plot histogram of time series as density then as a normal count.
        density, bins, patches = ax.hist(
            t,
            alpha=0.5,
            density=True,
            bins="auto",
            label="__nolabel__",
        )
        ax2.hist(
            t,
            facecolor="none",
            bins=bins,
            edgecolor="grey",
            linewidth=1,
        )

        # Plot pdf of distribution
        dc = get_dist(dist)(*params.isel(nbasins=i))
        mn = dc.ppf(0.01)
        mx = dc.ppf(0.99)
        q = np.linspace(mn, mx, 200)
        pdf = dc.pdf(q)

        ps = ", ".join([f"{x:.1f}" for x in p.values])
        ax.plot(q, pdf, "-", label=f"{params.attrs['scipy_dist']}({ps})")

        # Labels
        ax.set_xlabel(f"{ts.long_name} (${units2pint(ts.units):~P}$)")
        ax.set_ylabel("Probability density")
        ax2.set_ylabel("Histogram count")

        ax.legend(frameon=False)

    plt.tight_layout()
    return fig


def forecast(file: Union[str, Path], fcst_var: str = "q_sim") -> matplotlib.pyplot.Figure:
    """
    Create a graphic of the hydrograph for each forecast member.

    Parameters
    ----------
    file : str or Path
        Raven output file containing simulated streamflows.
    fcst_var : str
        Name of the streamflow variable.

    Returns
    -------
    matplotlib.pyplot.Figure
    """
    ds = xr.open_dataset(file)

    # Get time data for the plot
    dates = pd.DatetimeIndex(ds.time.values)
    start = dates.min()
    end = dates.max()

    fig, ax = plt.subplots()  # initialize figure

    # Plot the simulated streamflows for each hydrological model
    ds[fcst_var].plot.line("b", x="time", add_legend=False)

    # plt.xlim([first_date, last_date])
    ax.set_xlabel("Time")
    ax.set_ylabel("Streamflow (m³/s)")
    ax.set_title(f"Forecasted hydrograph between {start:%Y/%m/%d} and {end:%Y/%m/%d}.")
    ax.grid()

    plt.xticks(rotation=90)
    plt.tight_layout()

    return fig


def hindcast(file: Union[str, Path], fcst_var: str, qobs: Union[str, Path], qobs_var: str) -> matplotlib.pyplot.Figure:
    """
    Create a graphic of the hydrograph for each hindcast member.

    Parameters
    ----------
    file : str or Path
        Raven output file containing simulated streamflows.
    fcst_var : str
        Name of the streamflow variable.
    qobs : str or Path
        Streamflow observation file, with times matching the hindcast.
    qobs_var : str
        Name of the streamflow observation variable.

    Returns
    -------
    matplotlib.pyplot.Figure
    """
    ds = xr.open_dataset(file)
    ds2 = xr.open_dataset(qobs)

    # Get time data for the plot
    dates = pd.DatetimeIndex(ds.time.values)
    start = dates.min()
    end = dates.max()

    fig, ax = plt.subplots()  # initialize figure

    # Plot the simulated streamflows for each hydrological model
    hh = ds[fcst_var].plot.line("b", x="time", label="Hindcasts")
    ho = ds2[qobs_var].plot.line("r", label="Observations")

    # plt.xlim([first_date, last_date])
    ax.set_xlabel("Time")
    ax.set_ylabel("Streamflow (m³/s)")
    ax.set_title(f"Hindcasted hydrograph between {start:%Y/%m/%d} and {end:%Y/%m/%d}.")

    # Add legend
    handles = hh[:1] + ho
    ax.legend(handles, [h.get_label() for h in handles])
    ax.grid()

    plt.xticks(rotation=90)
    plt.tight_layout()

    return fig



from dataclasses import fields
from typing import Union

import numpy as np
import pint
import xarray as xr

from ravenpy.config.emulators import get_model


def realization(n):
    """
    Return a realization coordinate.

    Parameters
    ----------
    n : int
        Size of the ensemble.
    """
    return xr.IndexVariable(
        "members",
        data=range(n),
        attrs={
            "standard_name": "members",
            "axis": "E",
            "units": "1",
            "long_name": "Label identifying the ensemble member",
        },
    )


def param(model):
    """
    Return a parameter coordinate.

    Parameters
    ----------
    model : str
        Model name.
    """
    cls = get_model(model)
    P = cls.model_fields["params"].annotation  # noqa: N806
    return xr.IndexVariable(
        "param",
        data=np.array([f.name for f in fields(P)]),
        attrs={
            "standard_name": "parameter",
            "long_name": f"{model} model parameter name",
        },
    )


# FIXME: cumulative is not used
def infer_scale_and_offset(
    da: xr.DataArray,
    data_type: str,
    cumulative: bool = False,  # noqa: F841
) -> tuple[float, float]:
    """
    Return scale and offset parameters from data.

    Infer scale and offset parameters describing the linear transformation from the units in file to Raven
    compliant units.

    Parameters
    ----------
    da : xr.DataArray
        Input data.
    data_type : str
        Raven data type, e.g. 'PRECIP', 'TEMP_AVE', etc.
    cumulative : bool
        Default: False.

    Returns
    -------
    float, float
        Scale and offset parameters.

    Notes
    -----
    Does not work with accumulated variables.
    """
    import pint
    from xclim.core.units import FREQ_UNITS, parse_offset, units, units2pint

    from ravenpy.config import defaults

    # Read units attribute from netCDF file -- assuming CF-Compliance.
    source = units2pint(da.attrs["units"])

    # Get default units for data type.
    target = defaults.units[data_type]

    # Linear transform parameters
    try:
        scale, offset = units_transform(source, target)
    except pint.errors.DimensionalityError as err:
        if data_type in ["PRECIP", "PRECIP_DAILY_AVE", "RAINFALL", "SNOWFALL"]:
            # Source units are in total precipitation instead of rate. We need to infer the accumulation time
            # in order to find the transform.
            freq = xr.infer_freq(da.time)
            if freq is None:
                raise ValueError(f"Cannot infer time frequency of input data {da}") from err
            multi, base, start_anchor, _ = parse_offset(freq)
            if base in ["M", "Q", "A"]:
                raise ValueError(f"Irregular time frequency for input data {da}") from err
            real_source = source / multi / units(FREQ_UNITS.get(base, base))
            scale, offset = units_transform(real_source, target)
        else:
            raise NotImplementedError(f"data_type: {data_type}") from err

    return scale, offset


def units_transform(source: Union[str, pint.Unit], target: str, context: str = "hydro"):
    """
    Return linear transform parameters to convert one unit to another.

    If the target unit is given by `y = ax + b`, where `x` is the value of the source unit, then this function
    returns a, b.

    Parameters
    ----------
    source : str, pint.Unit
        Source unit string, pint-recognized.
    target : str
        Target unit string, pint-recognized.
    context : str
        Context of unit conversion. Default: "hydro".
    """
    from xclim.core.units import convert_units_to

    b = convert_units_to(0 * source, target, context)
    a = convert_units_to(1 * source, target, context) - b

    return round(a, 9), round(b, 9)



import pytest
from packaging.version import Version
from pydap import __version__ as __pydap_version__

from ravenpy.config.utils import nc_specs


older_pydap = False
if Version(__pydap_version__) < Version("3.5.5"):
    older_pydap = True


def test_nc_specs(yangtze):
    f = yangtze.fetch("raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc")
    attrs = nc_specs(f, "PRECIP", station_idx=1, alt_names=("rain",))
    assert "file_name_nc" in attrs


def test_nc_specs_bad(bad_netcdf):
    # Test with scalar elevation. Should normally have a station dimension, but not always the case.
    attrs_s = nc_specs(bad_netcdf, "PRECIP", station_idx=1, alt_names=("rain",))
    assert attrs_s["elevation"] == 1.0


@pytest.mark.online
@pytest.mark.skipif(older_pydap, reason="pydap version 3.5.5 is required for this test", strict=False)
def test_dap_specs():
    # Link to THREDDS Data Server netCDF testdata
    tds = "https://pavics.ouranos.ca/twitcher/ows/proxy/thredds/dodsC/birdhouse/testdata/raven"
    fn = f"{tds}/raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc"

    attrs = nc_specs(fn, "PRECIP", station_idx=1, alt_names=("rain",), engine="pydap")
    assert "units" in attrs



from collections.abc import Sequence
from dataclasses import field, make_dataclass
from typing import Union

from pydantic import Field, field_validator
from pydantic.dataclasses import dataclass
from pymbolic.primitives import Variable

import ravenpy.config.processes as p
from ravenpy.config import commands as rc
from ravenpy.config import options as o
from ravenpy.config.base import Sym, SymConfig
from ravenpy.config.commands import (
    HRU,
    PL,
    LandUseClasses,
    LandUseParameterList,
    SoilClasses,
    SoilParameterList,
    SoilProfiles,
    VegetationClasses,
)
from ravenpy.config.defaults import nc_attrs
from ravenpy.config.rvs import Config


P = dataclass(
    make_dataclass(
        "Params",
        [(f"X{i:02}", Sym, field(default=Variable(f"X{i:02}"))) for i in range(1, 36)]
        + [(f"R{i:02}", Sym, field(default=Variable(f"R{i:02}"))) for i in range(1, 9)],
    ),
    config=SymConfig,
)

# Bug: RavenC tries to write two variables with the same name `Snow Melt (Liquid) [mm]` to netCDF.


class LandHRU(HRU):
    land_use_class: str = "FOREST"
    veg_class: str = "FOREST"
    soil_profile: str = "DEFAULT_P"
    aquifer_profile: str = "[NONE]"
    terrain_class: str = "DEFAULT_T"


class HRUs(rc.HRUs):
    """
    HRUs command for GR4J.

    Pydantic is able to automatically detect if an HRU is Land or Lake if `hru_type` is provided.
    """

    root: Sequence[LandHRU]


class Blended(Config):
    """
    Blended hydrological model for blending outputs from different submodules.

    References
    ----------
    Mai, J., Craig, J. R., and Tolson, B. A.: Simultaneously determining global
    sensitivities of model parameters and model structure, Hydrol. Earth Syst.
    Sci., 24, 5835–5858, https://doi.org/10.5194/hess-24-5835-2020, 2020.

    Chlumsky, R., Mai, J., Craig, J. R., & Tolson, B. A. (2021). Simultaneous
    calibration of hydrologic model structure and parameters using a blended
    model. Water Resources Research, 57, e2020WR029229.
    https://doi.org/10.1029/2020WR029229
    """

    params: P = P()
    hrus: HRUs = Field(
        [
            LandHRU(),
        ],
        alias="HRUs",
    )
    netcdf_attribute: dict[str, str] = {"model_id": "Blended"}
    sub_basins: rc.SubBasins = Field([rc.SubBasin()], alias="SubBasins")
    write_netcdf_format: bool = Field(True, alias="WriteNetcdfFormat")
    time_step: Union[float, str] = Field(1.0, alias="TimeStep")
    calendar: o.Calendar = Field("PROLEPTIC_GREGORIAN", alias="Calendar")
    routing: o.Routing = Field("ROUTE_NONE", alias="Routing")
    catchment_route: o.CatchmentRoute = Field("ROUTE_DUMP", alias="CatchmentRouting")
    evaporation: o.Evaporation = Field(o.Evaporation.OUDIN, alias="Evaporation")
    rain_snow_fraction: o.RainSnowFraction = Field(o.RainSnowFraction.HBV, alias="RainSnowFraction")
    potential_melt_method: o.PotentialMeltMethod = Field(o.PotentialMeltMethod.HMETS, alias="PotentialMeltMethod")
    soil_model: rc.SoilModel = Field(3, alias="SoilModel")

    hydrologic_processes: Sequence[Union[rc.Process, p.ProcessGroup]] = Field(
        [
            p.Precipitation(algo="RAVEN_DEFAULT", source="ATMOS_PRECIP", to="MULTIPLE"),
            p.ProcessGroup(
                p=[
                    p.Infiltration(algo="INF_HMETS", source="PONDED_WATER", to="MULTIPLE"),
                    p.Infiltration(algo="INF_VIC_ARNO", source="PONDED_WATER", to="MULTIPLE"),
                    p.Infiltration(algo="INF_HBV", source="PONDED_WATER", to="MULTIPLE"),
                ],
                params=[P.R01, P.R02],
            ),
            p.Overflow(algo="OVERFLOW_RAVEN", source="SOIL[0]", to="CONVOLUTION[1]"),
            p.ProcessGroup(
                p=[
                    p.Baseflow(
                        algo="BASE_LINEAR_ANALYTIC",
                        source="SOIL[0]",
                        to="SURFACE_WATER",
                    ),
                    p.Baseflow(algo="BASE_VIC", source="SOIL[0]", to="SURFACE_WATER"),
                    p.Baseflow(algo="BASE_TOPMODEL", source="SOIL[0]", to="SURFACE_WATER"),
                ],
                params=(P.R03, P.R04),
            ),
            p.Percolation(algo="PERC_LINEAR", source="SOIL[0]", to="SOIL[1]"),
            p.Overflow(algo="OVERFLOW_RAVEN", source="SOIL[1]", to="CONVOLUTION[1]"),
            p.Percolation(algo="PERC_LINEAR", source="SOIL[1]", to="SOIL[2]"),
            p.ProcessGroup(
                p=[
                    p.SoilEvaporation(algo="SOILEVAP_ALL", source="SOIL[0]", to="ATMOSPHERE"),
                    p.SoilEvaporation(algo="SOILEVAP_TOPMODEL", source="SOIL[0]", to="ATMOSPHERE"),
                ],
                params=(P.R05,),
            ),
            p.Convolve(algo="CONVOL_GAMMA", source="CONVOLUTION[0]", to="SURFACE_WATER"),
            p.Convolve(algo="CONVOL_GAMMA_2", source="CONVOLUTION[1]", to="SURFACE_WATER"),
            p.ProcessGroup(
                p=[
                    p.Baseflow(
                        algo="BASE_LINEAR_ANALYTIC",
                        source="SOIL[1]",
                        to="SURFACE_WATER",
                    ),
                    p.Baseflow(algo="BASE_POWER_LAW", source="SOIL[1]", to="SURFACE_WATER"),
                ],
                params=[
                    P.R06,
                ],
            ),
            p.ProcessGroup(
                p=[
                    p.SnowBalance(algo="SNOBAL_HMETS", source="MULTIPLE", to="MULTIPLE"),
                    p.SnowBalance(algo="SNOBAL_SIMPLE_MELT", source="SNOW", to="PONDED_WATER"),
                    p.SnowBalance(algo="SNOBAL_HBV", source="MULTIPLE", to="MULTIPLE"),
                ],
                params=(P.R07, P.R08),
            ),
        ],
        alias="HydrologicProcesses",
    )

    soil_classes: SoilClasses = Field(
        [
            {"name": "TOPSOIL"},
            {"name": "PHREATIC"},
            {"name": "DEEP_GW"},
        ],
        alias="SoilClasses",
    )
    vegetation_classes: VegetationClasses = Field(
        [{"name": "FOREST", "max_ht": 4, "max_lai": 5, "max_leaf_cond": 5}],
        alias="VegetationClasses",
    )
    land_use_classes: LandUseClasses = Field(
        [rc.LU(name="FOREST", impermeable_frac=0, forest_coverage=0.02345)],
        alias="LandUseClasses",
    )
    terrain_classes: rc.TerrainClasses = Field(
        [
            rc.TC(
                name="DEFAULT_T",
                hillslope_length=1,
                drainage_density=1,
                topmodel_lambda=P.X07,
            )
        ],
        alias="TerrainClasses",
    )

    soil_profiles: SoilProfiles = Field(
        [
            {"name": "LAKE"},
            {"name": "ROCK"},
            {
                "name": "DEFAULT_P",
                "soil_classes": ("TOPSOIL", "PHREATIC", "DEEP_GW"),
                "thicknesses": (P.X29, P.X30, 1e6),
            },
        ],
        alias="SoilProfiles",
    )

    global_parameter: dict = Field(
        {
            "SNOW_SWI_MIN": P.X13,
            "SNOW_SWI_MAX": P.X14,
            "SWI_REDUCT_COEFF": P.X15,
            "SNOW_SWI": P.X19,
            "RAINSNOW_TEMP": P.X31,
            "RAINSNOW_DELTA": P.X32,
        }
    )

    soil_parameter_list: SoilParameterList = Field(
        {
            "parameters": (
                "POROSITY",
                "PERC_COEFF",
                "PET_CORRECTION",
                "BASEFLOW_COEFF",
                "B_EXP",
                "HBV_BETA",
                "MAX_BASEFLOW_RATE",
                "BASEFLOW_N",
                "FIELD_CAPACITY",
                "SAT_WILT",
            ),
            "pl": [
                PL(
                    name="TOPSOIL",
                    values=(
                        1.0,
                        P.X28,
                        P.X08,
                        P.X04,
                        P.X02,
                        P.X03,
                        P.X05,
                        P.X06,
                        P.X10,
                        P.X09,
                    ),
                ),
                PL(
                    name="PHREATIC",
                    values=(1.0, P.X35, 0, P.X11, 0, 0, 0, P.X12, 0, 0),
                ),
                PL(
                    name="DEEP_GW",
                    values=(1.0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
                ),
            ],
        },
        alias="SoilParameterList",
    )
    land_use_parameter_list: LandUseParameterList = Field(
        {
            "parameters": (
                "MIN_MELT_FACTOR",
                "MAX_MELT_FACTOR",
                "DD_MELT_TEMP",
                "DD_AGGRADATION",
                "REFREEZE_FACTOR",
                "REFREEZE_EXP",
                "DD_REFREEZE_TEMP",
                "HMETS_RUNOFF_COEFF",
                "GAMMA_SHAPE",
                "GAMMA_SCALE",
                "GAMMA_SHAPE2",
                "GAMMA_SCALE2",
                "FOREST_SPARSENESS",
            ),
            "pl": [
                PL(
                    name="[DEFAULT]",
                    values=(
                        P.X24,
                        P.X25,
                        P.X26,
                        P.X27,
                        P.X18,
                        P.X17,
                        P.X16,
                        P.X01,
                        P.X20,
                        P.X21,
                        P.X22,
                        P.X23,
                        0,
                    ),
                )
            ],
        },
        alias="LandUseParameterList",
    )
    vegetation_parameter_list: rc.VegetationParameterList = Field(
        {
            "parameters": ("RAIN_ICEPT_PCT", "SNOW_ICEPT_PCT", "SAI_HT_RATIO"),
            "pl": [PL(name="[DEFAULT]", values=(0, 0, 0))],
        },
        alias="VegetationParameterList",
    )
    seasonal_relative_lai: rc.SeasonalRelativeLAI = Field(rc.SeasonalRelativeLAI(), alias="SeasonalRelativeLAI")
    seasonal_relative_height: rc.SeasonalRelativeHeight = Field(rc.SeasonalRelativeHeight(), alias="SeasonalRelativeHeight")

    hru_state_variable_table: rc.HRUStateVariableTable = Field(
        [
            rc.HRUState(
                hru_id=1,
                data={
                    "SOIL[0]": P.X29 * 500,
                    "SOIL[1]": P.X30 * 500,
                },
            ),
        ],
        alias="HRUStateVariableTable",
    )
    _nc_attrs = field_validator("netcdf_attribute")(nc_attrs)

    def __init__(self, **data):
        super().__init__(**data)

        if self.gauge:
            for gauge in self.gauge:
                gauge.rain_correction = self.params.X33
                gauge.snow_correction = self.params.X34



from ravenpy import __raven_version__


units = {
    "PRECIP": "mm/d",
    "PRECIP_DAILY_AVE": "mm/d",
    "PRECIP_5DAY": "mm",
    "SNOW_FRAC": "",
    "SNOWFALL": "mm/d",
    "RAINFALL": "mm/d",
    "RECHARGE": "mm/d",
    "TEMP_AVE": "degC",
    "TEMP_DAILY_AVE": "degC",
    "TEMP_MIN": "degC",
    "TEMP_DAILY_MIN": "degC",
    "TEMP_MAX": "degC",
    "TEMP_DAILY_MAX": "degC",
    "TEMP_MONTH_MAX": "degC",
    "TEMP_MONTH_MIN": "degC",
    "TEMP_MONTH_AVE": "degC",
    "TEMP_AVE_UNC": "degC",
    "TEMP_MAX_UNC": "degC",
    "TEMP_MIN_UNC": "degC",
    "AIR_DENS": "kg/m**3",
    "AIR_PRES": "kPa",
    "REL_HUMIDITY": "",
    "ET_RADIA": "MJ/m**2/d",
    "SHORTWAVE": "MJ/m**2/d",
    "SW_RADIA": "MJ/m**2/d",
    "SW_RADIA_NET": "MJ/m**2/d",
    "LW_RADIA_NET": "MJ/m**2/d",
    "LW_INCOMING": "MJ/m**2/d",
    "CLOUD_COVER": "",
    "DAY_LENGTH": "d",
    "DAY_ANGLE": "",
    "WIND_VEL": "m/s",
    "PET": "mm/d",
    "OW_PET": "mm/d",
    "PET_MONTH_AVE": "mm/d",
    "POTENTIAL_MELT": "mm/d",
    "SUBDAILY_CORR": "",
    "HYDROGRAPH": "m**3/s",
}

RAVEN_NO_DATA_VALUE = -1.2345
CALENDAR = "PROLEPTIC_GREGORIAN"


def nc_attrs(cls, val):
    """Ensure default netCDF attributes are present."""
    if "model_id" not in val:
        raise ValueError("The key 'model_id' must be present in the input dictionary.")

    out = default_nc_attrs()
    out.update(val)
    return out


def default_nc_attrs():
    """Return default NetCDF global attributes."""
    import datetime as dt

    now = dt.datetime.now().isoformat(timespec="seconds")
    version = __raven_version__

    return {
        "history": f"Created on {now} by Raven {version}",
        "references": "Craig, J.R., and the Raven Development Team, Raven user's and developer's manual "
        f"(Version {version}), URL: https://raven.uwaterloo.ca/ ({dt.datetime.today().year}).",
    }



from ravenpy.utilities.coords import infer_scale_and_offset


def test_infer_scale_and_offset(open_dataset):
    # ERA5 precip and tas
    ts = "era5/tas_pr_20180101-20180108.nc"

    with open_dataset(ts) as ds:
        p = infer_scale_and_offset(ds["pr"], "PRECIP")
        assert p == (24000, 0)

        p = infer_scale_and_offset(ds["tas"], "TEMP_DAILY_AVE")
        assert p == (1, -273.15)

    # CMIP precip in kg / m^2 / s
    ts2 = "cmip5/nasa_nex-gddp-1.0_day_inmcm4_historical+rcp45_nex-gddp_1971-1972_subset.nc"

    with open_dataset(ts2) as ds:
        p = infer_scale_and_offset(ds["pr"], "PRECIP")
        assert p == (86400, 0)

    # VIC streamflow inputs in mm accumulated over 6 hours
    fn = "raven-routing-sample/VIC_streaminputs.nc"

    with open_dataset(fn) as ds:
        p = infer_scale_and_offset(ds.Streaminputs, "PRECIP")
        assert p == (4, 0)

    # GEPS dataset in cumulative m fails
    # fn = "caspar_eccc_hindcasts/geps_watershed.nc"
    # with open_dataset(fn) as ds:
    #     p = infer_scale_and_offset(ds.pr, "PRECIP")
    #     assert p == (1000, 0)
